
from attune.src.qtattune import run_app
import subprocess

import sys
import os


class Gui:
    """
    The attune-gui for usage within Python
    """

    def __init__(self):
        pass

    def run(self):
        # runs the attune gui within Python

        run_app()


def run_gui():
    g = Gui()
    g.run()


if __name__ == "__main__":
    run_gui()



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod

"""
EXPERIMENTAL and not tested!

This ships a functionality to analytically solve a CGE model using a
'production tree' approach. BETA has not yet been fully tested
"""

from sympy import *
import sys


def construct_prodfunc(entry,product,factors,learning):
    """

    EXPERIMENTAL and not tested!

    constructs a linear production function
    :param entry:  'linear' only one allowed currently
    :param product: str, what is produced by this agent?
    :param factors: list of str (length 2) what are the input factors?
    :param learning: learning: technological learning dict {product:learnin_coefficient}, i.e. is the input modified
    """

    if entry["type"] == "linear":
        coeffs = entry["coeffs"]

        symcoeffs = Matrix(len(coeffs),1,[Symbol(i) for i in coeffs])
        symfactors = Matrix(len(factors), 1, [Symbol(i) for i in factors])

        equation = sum([symcoeffs[i]* symfactors[i] for i in range(len(coeffs))]) - Symbol(product)
        if str(product) in learning:
            equation -= learning[str(product)]

        cond = str(sum(symcoeffs)) + "= 1"
        parameters = sum(symcoeffs) - 1

        unknowns = [Symbol(i) for i in factors] + [Symbol(product)]
        # unknowns += [i for i in symcoeffs]

        #print(equation)
        return equation,cond,unknowns,parameters


    else:
        raise NotImplementedError("Other than type 'linear' currently not possible.")



def construct_costfunc(entry,product,factors,learning):
    if entry["type"] == "linear":
        coeffs = entry["coeffs"]

        symcoeffs = Matrix(len(coeffs),1,[Symbol(i) for i in coeffs])
        symprices = Matrix(len(factors), 1, [Symbol("p_"+i) for i in factors])

        equation = sum([symcoeffs[i]* symprices[i] for i in range(len(coeffs))]) - Symbol("p_"+product)
        if str(product) in learning:
            equation += learning[str(product)]

        cond = str(sum(symcoeffs)) + "=1"
        parameters = sum(symcoeffs) - 1

        unknowns = [Symbol(i) for i in factors] + [Symbol(product)]
        # unknowns += [i for i in symcoeffs]

        return equation, cond, unknowns, parameters

    else:
        raise NotImplementedError("Other than type 'linear' currently not possible.")


def nextwait(s):
    d = {
        "/":"- ",
        "- ":"\\",
        "\\":"|",
        "|":"+",
        "+":"/"
    }
    return d[s]


def traverse(t, production_dict,learning, l=None,s = None,unknowns = None, abbrevs=None,topnode=None,wait="/",rec_depth=0):
    """
    EXPERIMENTAL and not tested!

    traverses a tree
    :param t:  tree (nested dict)
    :param production_dict: production dict
    :param learning: dict storing improvements in technology

    :param l: recursion param, do not use , stores equations
    :param s: recursion param, do not use, stores equations for elasticities
    :param unknowns: recursion paramm, do not use
    :param topnode:  recursion param, do not use
    :param rec_depth: recursion depth
    :return:
    """

    # TODO make this more efficient

    if l is None: # stores equations
        l = []
    if unknowns is None: # stores unknowns
        unknowns = []
    if s is None:  # stores the equations for elasticities
        s = []
    if abbrevs is None: # stores abbreviations
        abbrevs = []

    # generate production and cost functions

    if topnode is not None:
        topnode_s = topnode.split(";")[0]
        unknowns.append(topnode_s)
        unknowns.append("p_" + topnode_s)

    for k,v in t.items(): # traverse sub-tree

        # has production parameters
        splitk = k.split(";")
        print("   "*rec_depth,"|__ visit",splitk[0],"    "*10)

        fact = splitk[0].strip()
        unknowns.append(fact)
        unknowns.append("p_"+fact)

        # production

        if len(splitk) > 1:  # has production parameters
            fact = splitk[0].strip()
            factors = [v.split(";")[0].strip() for v in v.keys()]

            prodfunc,cond1,unknowns1,parameters = construct_prodfunc(
                entry = production_dict[splitk[1].strip()],
                product = fact,
                factors = factors,
                learning=learning,
                )

            costfunc, cond2, unknowns2, parameters = construct_costfunc(
                entry=production_dict[splitk[1].strip()],
                product=fact,
                factors=factors,
                learning=learning)

            l.append(prodfunc)
            l.append(costfunc)

            abbrevs += [parameters]  # str(cond1))
            # abbrevs.append(parameters)  # str(cond2))

            for f in unknowns1+unknowns2:

                unknowns.append(f)
                #unknowns.append(Symbol("p_"+str(f)))

            if v != {} and len(v) > 1:
                print("   "*rec_depth," ___", " %i equations" % len(list(set(l))) )
                for l_i in list(set(l)):
                    print("   " * (1+rec_depth), "    ",l_i)

                new_vals = traverse(v, production_dict, learning, l.copy(), s.copy(), topnode=k, abbrevs=abbrevs, unknowns=unknowns, wait=wait, rec_depth=rec_depth+1)
                l += new_vals[0]
                unknowns += new_vals[1]

        else:  # has no production parameters

            if v != {}:
                if len(v.keys()) >= 1:
                    print("   "*rec_depth," ___", " %i equations" % len(list(set(l))) )
                    for l_i in list(set(l)):
                        print("   " * (1 + rec_depth), "    ", l_i)

                    new_vals = traverse(v, production_dict, learning, l.copy(), s.copy(), topnode=k, abbrevs=abbrevs, unknowns=unknowns, wait=wait, rec_depth=rec_depth+1)
                    l+= new_vals[0]
                    unknowns+= new_vals[1]

                if len(v.keys()) == 1:
                    vv = list(iter(v))[0].split(";")[0]

                    eq1 = Symbol(k)-Symbol(vv)
                    eq2 = Symbol("p_"+k)-Symbol("p_"+vv)

                    if str(k) in learning:
                        #print(str(k), "IN LEARNING")

                        eq1 -= Symbol(learning[k])
                        eq2 += Symbol(learning[k])

                    elif str(vv) in learning:
                        #print(str(vv), "IN LEARNING")

                        eq1 += Symbol(learning[vv])
                        eq2 -= Symbol(learning[vv])

                    #print("eq1",eq1)
                    #print("eq2",eq2)

                    l.append(eq1)
                    l.append(eq2)

        # generate elasticities
        if len(v.keys()) == 2:

            keys =  list(v.keys())

            k0 = keys[0].split(";")[0]
            k1 = keys[1].split(";")[0]
            Y1 = Symbol(k0)
            Y2 = Symbol(k1)

            p1 = Symbol("p_"+k0)
            p2 = Symbol("p_"+k1)

            k0 = k.split(";")[0].strip()
            k1 = k.split(";")[1].strip()
            sigma = "sigma_" + k0

            if k1 in production_dict:
                if "elast" in production_dict[k1]:
                    sigma = production_dict[k1]["elast"]

            elast = Symbol(sigma)

            s_eq = (Y1-Y2) - elast * (p2-p1)
            l.append(s_eq)
            print("   " * rec_depth, "    ", s_eq)

    else:
        pass

        # TODO think of a way to combine elasticities for more than two factors(?)


    unknowns = [str(u) for u in unknowns]

    print("   "*rec_depth,"< RETURN")
    return l, list(set(unknowns)), list(set(s)),list(set(abbrevs))


def make_list(x):
    if isinstance(x, list):
        return x
    else:
        return [x]


def apply_simplifications(expr,simplifications,wait="/",silent=False):
    """
    EXPERIMENTAL and not tested!
    """

    out = []

    for e in make_list(expr):

        if not silent:
            wait = nextwait(wait)
            sys.stdout.write("  " + wait + "\r")

        e2 = e

        for i, simpl in enumerate(simplifications):

            s = simpl.split("=")
            s1 = Symbol(s[0].strip())

            ss1 = s[1].strip()

            if ss1 == "0":
                s2 = 0
            else:
                s2 = Symbol(str(ss1))

            if isinstance(e2,str):
                e2 = Symbol(e2)

            # if s[0].strip() == "p_V_X":
            #    print(str(s1),"->",str(s2),"                                    ")

            e2 = e2.subs(s1,s2,evaluate=True).simplify()

        # out.append(e2)
        if str(e2) != "0":
            out.append(e2)
            if not silent:
                sys.stdout.write( " %5s" %" " + str(e2)+"%50s\r\r"%" "+"\r")

        # print(" > ")

    # print("OUT",out)
    return list(set(out))


def gen_equation_system(t_prod, t_markets, production_dict, simplifications, learning):
    """
    EXPERIMENTAL and not tested!

    :param t_prod: dict or list of dicts of production
    :param t_markets: dict or list of dicts of markets (sector aggregation)
    :param production_dict:
    :return:
    """

    equation_system = []
    unknown_system = []

    sys.stdout.write("Generating equation system...\n")

    # traverse

    unknowns = None
    lst = None
    s0 = None

    abbr = []

    for t in make_list(t_prod):

        lst,unknowns,s0,abbrevs = traverse(t, production_dict,learning)
        abbr += [a for a in abbrevs]

        # [print(l) for l in lst]  # print list of equations

        print("Juggling equations...")
        u = apply_simplifications(unknowns,simplifications)
        l = apply_simplifications(lst,simplifications)
        s = apply_simplifications(s0, simplifications)

        equation_system += l
        equation_system += s
        unknown_system += u

    for t in make_list(t_markets):

        lst,unknowns,s0,abbrevs = traverse(t, production_dict,learning)

        abbr += [a for a in abbrevs]

        print("Juggling equations...")
        u = apply_simplifications(unknowns,simplifications)
        l = apply_simplifications(lst,simplifications)
        s = apply_simplifications(s0, simplifications)

        equation_system += l
        equation_system += s
        unknown_system += u

    sys.stdout.write("done! %50s\n\n"%" ")

    # print("ABBREVS",abbr)

    return list(set(equation_system)), list(set(unknown_system)) + list(set(list(learning.values()))),list(set(abbr))



def convert_symbols(x):
    y = []

    for xi in x:
        if isinstance(xi,str):
            y.append(Symbol(xi))
        else:
            y.append(xi)

    return y

class CGEModel:

    def __init__(self, t_prod, t_markets, production_dict, simplifications,learning):
        """

        EXPERIMENTAL and not tested!

        :param t_prod: dict or list of dicts of production
        :param t_markets: dict or list of dicts of markets (sector aggregation)
        :param production_dict:
        :param simplifications:

        e.g.

        # consumption sector
        t_prod = {
            "C": {
                "X;PFX": {
                    "V_X": {},
                    "S_X": {"E_X": {"V_E": {}}},
                },

            }
        }

        # value added sector
        t2= {
            "V;PFV": {
                "V_X": {},
                #"V_Y": {},
                "V_E": {}
            }
        }
        t_markets = [t2]

        # production parameters
        production_dict = {
            "PFC": {"type": "linear", "coeffs": ["theta_XC"]},
            "PFX": {"type": "linear", "coeffs": ["theta_XV", "theta_XS"]},

            "PFV": {"type": "linear", "coeffs": ["alpha_X", "alpha_E"]},
        }

        # simplifications
        simplifications = [
            "p_E_X = p_E",
            "p_V_X = p_V",
            "E_X = E",
            "p_V_E = p_V",
            "p_V = 0",
            "V = 0",
        ]

        # technological learning parameters
        learning = {
            "S_X": "A_X"
        }

        """

        eq_sys, unknowns, ident = gen_equation_system(t_prod, t_markets, production_dict, simplifications,learning)

        self.equations = eq_sys

        self.unknowns = convert_symbols(unknowns)   # + list(set(list(learning.values())))
        self.simplifications = simplifications
        self.parameter_identities = ident

    def __str__(self):
        return "<CGEMODEL[%i] with %i equations and %i unknowns>" % (id(self),len(self.equations),len(self.unknowns))

    def apply_simplifications(self,expr,silent=False):
        # expr: sympy expression or list of sympy expressions to simplify
        # :returns: simplified expr as list

        simple_expr = apply_simplifications(expr, self.simplifications,silent=silent)
        if len( simple_expr) == 1:
            return simple_expr[0]
        else:
            return simple_expr


    def printer(self):

         print("%i EQUATIONS\n"%(len(self.equations)))

         for k, i in enumerate(self.equations):
            #print("[%02i]" % k)
            sys.stdout.write("[%3i]  "%k)
            pprint(i,wrap_line =False)
            #print("\n")

         print("\n\n %i UNKNOWNS\n"%(len(self.unknowns)))

         for k, i in enumerate(self.unknowns):
                sys.stdout.write("[%3i]  " % k)
                #print("[%02i]" % k)
                pprint(i)

         print("\n\n %i IDENTITIES\n" % (len(self.parameter_identities)))

         for k, i in enumerate(self.parameter_identities):
             sys.stdout.write("[%3i]  " % k)
             # print("[%02i]" % k)
             pprint(i)

         print("\n")


    def filter_unknowns(self, equation_indices):

        equations = [self.equations[i] for i in equation_indices]

        u_selection = []

        for u in self.unknowns:

            for equation in equations:

                if isinstance(u,str):
                    u = Symbol(u)

                if u in equation.free_symbols:
                    u_selection.append(u)
                    break

        return u_selection



import numpy as np
import pandas 
from datetime import datetime
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt


def convert_quarterly_to_datetime(df):
    """
    converts a dataframe with a string index of form '2020-Q2' (typical Eurostat format)
    to a dataframe with a datetime index 

    :param df: a pandas dataframe
    :return: df_converted, a pandas dataframe with datetime-formatted index
    """
    # https://stackoverflow.com/questions/53898482/clean-way-to-convert-quarterly-periods-to-datetime-in-pandas
    df.index = pandas.PeriodIndex(df.index, freq='Q').to_timestamp()

    return df 


def convert_numeric(df):
    """
    converts all column of a dataframe 
    
    :param df: a pandas dataframe
    :return: converted dataframe
    """

    for column in df.columns:
        df[column] = pandas.to_numeric(df[column])

    print("df\n",df)
    return df 



def stretch_to_length(df,new_length,method="cubic"):
    """
    stretches a dataframe with datetime index to a new, higher length (by interpolation of the values).
    
    :param df: a pandas data frame with datetime-formatted index
    :param new_length: new desired length 
    :param method: intrepolation style (default 'cubic')

    :return df_new: the stretched dataframe
    """
    factor = (new_length/(len(df)))
    assert factor > 1.0, "new_length smaller than original length. aborting."
    df_new = stretch_pandas(df, factor, kind=method)
    assert len(df_new) == new_length, "internal error."
    return df_new 


def stretch_datetime(x, factor):
    """
    stretches a series or array containing datetime objects by a certain factor

    :param x: an array 
    :param factor: int, factor > 1 by which to stretch the data
    """
    x = np.array(x)

    assert type(x[0]) == datetime.date or np.issubdtype(x.dtype, np.datetime64), "Data index does not seem to be datetime. It is %s" % type(x[0])
    assert factor >= 1.0, "streching by less than factor 1.0 impossible."

    dt = max(x)-min(x)
    sign = 1
    if x[1] < x[0]:
        sign = -1

    ds = sign * dt/((len(x))*factor -1)

    new_ar = []

    if ds >= 0:
        #print("dt >=0")
        s = min(x) # x[0]

        while s <= max(x): #[-1]:
            #print("s", s)
            new_ar.append(s)
            s += ds

    elif ds < 0:
        #print("dt< 0","max,min",max(x),min(x))
        s = max(x)

        while s >= min(x):
            new_ar.append(s)
            #print("s", s)
            s += ds

    #print("stetch")
    #print(len(x))
    #print(len(new_ar))
    return pandas.Series(new_ar)


def interpolate(y,stretch_factor,kind="cubic"):
    """
    interpolates an array using scipy interp1d. 

    x is datetime, y is float or somehow numerical
    stretch_factor = 3 # quarter to month
    """
    x = np.arange(len(y))
    f = interp1d(x, y, kind=kind)
    x_new = np.linspace(0, len(y)-1, int(stretch_factor*len(y)))
    y_new = f(x_new)

    return y_new


def stretch_pandas(s, factor, kind="cubic"):
    """
    stretch a pandas dataframe or series by a certain factor 

    :param s: pandas dataframe or series
    :param factor: factor by which you want to stretch

    """
    try:
        s = s.to_frame()
    except:
        pass

    t = s.index # should be datetime
    indexname = s.index.name
    t2 = stretch_datetime(t, factor)

    #print("stretched time series",len(t2))

    new_data = {indexname:t2}
    for column in s.columns:
        new_data[column] = interpolate(s[column],factor,kind=kind)

    df =  pandas.DataFrame(new_data).set_index(indexname)
    #print("new df")
    return df


def log_difference(x0):
    """
    computes difference of log[x0], i.e. log[x1/x0] of a pandas time series
    """
    x = np.array(x0,dtype=float)

    y = np.zeros(len(x)-1)

    for i in range(len(x)-1):
        y[i] = np.log(x[i+1]) - np.log(x[i])

    z =  y - np.mean(y)
    z /= max(z)

    return z # pandas.Series(z,index=x0.index[:-1])
   

def difference(x0):
    """
    computes difference of x0, i.e. [x1-x0] of a pandas time series
    """
    x = np.array(x0,dtype=float)

    y = np.zeros(len(x)-1)

    for i in range(len(x)-1):
        y[i] = x[i+1] - x[i] # np.log(x[i+1]) - np.log(x[i])

    z = y - np.mean(y)
    z /= max(z)

    return pandas.Series(z,index=x0.index[:-1])


def percentage_change(x0):
    """
    computes difference of x0, i.e. [x1-x0] of a pandas time series
    """
    x = np.array(x0,dtype=float)

    y = np.zeros(len(x)-1)

    for i in range(len(x)-1):
        if x[i] != 0:
            y[i] = 100*(x[i+1] - x[i]) / x[i] # np.log(x[i+1]) - np.log(x[i])
        else:
            y[i] = np.inf

    return pandas.Series(y,index=x0.index[:-1])


def cross_correlate(x1,x2,lags=15):
    """
    computes the cross-correlation of x1 and x2 using plt.xcorr
    returns dataframe with xcorr data
    """
    # res = plt.xcorr(x1, x2 , plt.gca(), color="black", maxlags=lags,)
    # lag_data = res[0]
    # corr_data = res[1]

    x1 -= np.mean(x1)
    x2 -= np.mean(x2)

    x1 /= np.std(x1)
    x2 /= np.std(x2)

    corr_data = np.correlate(x1, x2, 'full' ,lags=lags)
    lag_data = np.arange(-len(x1) + 1, len(x1))
    corr_data = np.divide(corr_data,(len(x1)))

    res_df = pandas.DataFrame({"Lag":lag_data,"Cross-Correlation":corr_data}).set_index("Lag")
    # print(res_df.loc[[1,5],:])

    return res_df


def cross_correlate_plot(x1,x2,normalize=True,ax=None,show_plot=True,ylabel="Cross-Correlation",lags=15,color="black"):
    """
    generates a cross-correlation plot of x1 and x2

    :param lags: maxium number of lags to show
    :param ylabel: 
    :param normalize:
    :param color: color of the plot 
    """
    if ax is None:
        fig = plt.figure(figsize=(4,3))
        ax = fig.add_subplot(111)
    else:
        ax = plt.gca()
        fig = plt.gcf()

    try:
        plt.style.use("macro")
    except:
        pass

    ax.set_xlabel("Lag")
    ax.set_ylabel(ylabel)

    x1 = np.array(x1)
    x2 = np.array(x2)

    lags2 = range(-len(x1)+1, len(x2))
    cs = []
    for lag in lags2:
        idx_lower_x1 = max(lag, 0)
        idx_lower_x2 = max(-lag, 0)
        idx_upper_x1 = min(len(x1), len(x1)+lag)
        idx_upper_x2 = min(len(x2), len(x2)-lag)
        b1 = x1[idx_lower_x1:idx_upper_x1]
        b2 = x2[idx_lower_x2:idx_upper_x2]
        c = np.correlate(b1, b2)[0]
        if normalize:
            c = c / np.sqrt((b1**2).sum() * (b2**2).sum())
        cs.append(c)

    # plt.xcorr(x1, x2, ax, color=color, maxlags=lags)
    corrCoef = np.array(cs) # np.correlate(x1, x2, 'full') # ,lags=lags)
    lagsOut = np.arange(-len(x1) + 1, len(x1))

    ax.bar(lagsOut, corrCoef,color=color,width=0.4)

    plt.tight_layout()
    ax.set_xlim([-lags,lags])

    if show_plot:
        plt.show()
    


from sfctools import convert_quarterly_to_datetime, stretch_to_length, convert_numeric
from sfctools import Settings
import pandas as pd 
import matplotlib.pyplot as plt
import numpy as np 

"""
Small data preprocessing pipeline
"""

Settings().read_from_yaml("../settings.yml")
T = Settings().get_hyperparameter("T") # simulation length

"""
GDP data pre-process
"""
gdp_df = pd.read_csv("raw/tipsau20_gdp.csv",delimiter=";",skiprows=1) # .to_numeric()
gdp_df.index = gdp_df["Quarter"]
del gdp_df["Quarter"]
del gdp_df["GDP"]

gdp_df = gdp_df.iloc[1:,:]
gdp_df = convert_numeric(gdp_df)

# 1. convert quarterly to datetime
gdp_df = convert_quarterly_to_datetime(gdp_df) 

# 2. stretch data to simulation length
gdp_df = stretch_to_length(gdp_df,T,method="linear")

# 3. save 
arr = gdp_df["Percentage Change"].to_numpy()
np.savetxt("processed/tipsau20_gdp_processed.txt",arr)



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod


import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
from networkx.convert_matrix import to_pandas_adjacency,to_pandas_edgelist
from networkx.algorithms.traversal.beamsearch import bfs_beam_edges
from collections.abc import Mapping
from networkx.drawing.nx_agraph import graphviz_layout

def graph_from_dict(graph_data):
    """
    take a nested dict and convert it to a networkx graph
    """

    G = nx.DiGraph()  # empty directed graph

    q = list(graph_data.items())
    while q: # iterate through the layers
        v, d = q.pop()
        for nv, nd in d.items():
            G.add_edge(v, nv)
            if isinstance(nd, Mapping):
                q.append((nv, nd))
    return G


class TreeStruct:
    """
    wrapper for a special networkx graph

    tree structure (for supply chain processing)
    """

    def __init__(self, data, handler):
        """
        inits a static production tree from nested dict.

        :param data: data in form of a nested dict
        :param handler: handler to be called at each node

        example data

        data = {
            A: {
                B: {},
                C:
                {
                    D: {},
                    E: {}
                }
            }
        }

        for a tree like

                     A
                    / \
                   B   C
                      / \
                     D   E
        """

        self.tree = graph_from_dict(data)
        self.handler = handler

    def traverse(self):
        """
        traverse the tree struct and execute production nodes
        on the way. -> Breadth-first search
        """

        if self.handler is not None:

            for u, v in self.tree.edges:  # iterate through all edges and trigger handlers
                v.trigger(self.handler, u)

    def plot(self):
        """
        plots the production tree with networkx
        """

        plt.figure(figsize=(7, 5))
        pos = graphviz_layout(self.tree, prog='dot')
        nx.draw(self.tree, pos, with_labels=True, arrows=True, node_color='#d1ffdf')
        plt.show()


if __name__ == "__main__":
    """
    test for plotting
    """

    data = {
            "A": {
                "B": {},
                "C":
                {
                    "D": {},
                    "E": {}
                }
            }
        }

    ts = TreeStruct(data, "Produce")

    ts.plot()



from PyQt5 import QtWidgets, uic,QtGui
from PyQt5.QtCore import QUrl
from PyQt5.QtWidgets import QFileDialog,QMessageBox,QTableWidgetItem,QAbstractItemView,QListWidgetItem,QShortcut
import sys
import os
import yaml
import shutil
import pyperclip

# from pandasmodel import PandasModel
from PyQt5.QtGui import QDesktopServices
import pickle as pkl
import re
import matplotlib.pyplot as plt
from PyQt5.QtCore import Qt
from PyQt5 import QtCore
from PyQt5.QtGui import QKeySequence

from .mamba_interpreter2 import convert_code

from PyQt5.QtGui import QFont, QFontDatabase, QColor #, QSyntaxHighlighter, QTextCharFormat
from PyQt5.QtGui import QKeySequence,QFontMetrics

import numpy as np
from sfctools.misc.timeseries import cross_correlate



class OutputDisplay(QtWidgets.QDialog):
    # !!! 
    # EXPERIMENTAL AND GRAYED OUT FOR NOW
    # !!!

    def notify(self,message,title=" "):
        msg = QMessageBox(self)
        msg.setWindowTitle(title)
        msg.setText(message)
        msg.open()


    def __init__(self,parent,path=""):

        super(OutputDisplay, self).__init__(parent) # Call the inherited classes __init__ method
        ui_path = os.path.dirname(os.path.abspath(__file__))
        uic.loadUi(os.path.join(ui_path,'output_display.ui'), self) # Load the .ui file

        self.data_idx = 0
        self.data = None

        self.lock_idx = None

        self.data_range_max = None
        self.data_range_min = None

        self.Btn_copy.pressed.connect(self.copy_data)
        self.Btn_copy_2.pressed.connect(self.copy_data2)

        self.path = path
        self.path_label.setText(self.path)
        self.data_name_label.setText(" ")
        self.data_name_label2.setText(" ")

        self.Btn_update.pressed.connect(self.load_file)
        self.Btn_prev.pressed.connect(self.decrease_idx)
        self.Btn_next.pressed.connect(self.increase_idx)

        self.lock_slide.valueChanged.connect(self.toggle_lock)

        self.range_from.textChanged.connect(self.update_range)
        self.range_to.textChanged.connect(self.update_range)

        self.idx_from_edit.textChanged.connect(self.update_idx)
        self.idx_to_edit.textChanged.connect(self.update_idx)

        self.table.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.table.selectRow(self.data_idx)
        self.table.cellClicked.connect(self.update_idx)

        self.plot_data = {"data":None, "lock_data":None,"mydat":None} # store plot data so it does not need to be computed every iteration
        self.plot_rep = 15 # every N frames, the plots are updated
        self.plot_counter  = self.plot_rep # counts down

        # self.load_file()

        self.show()

    def copy_data(self):
        # copy current data to clipboard
        if self.data is not None:

            try:

                v = self.data[list(self.data.keys())[self.data_idx]]
                if isinstance(v,dict):
                    v = list(v.values())

                title = list(self.data.keys())[self.data_idx]

                pyperclip.copy(title + "\n" + "\n".join([str(vi) for vi in v]))
            except Exception as e:
                self.notify(str(e),title="Error")

    def copy_data2(self):
        # copy current data to clipboard
        if self.data is not None:

            try:

                if self.lock_idx is not None: # cross-correlation
                    lock_data = self.data_from_idx(self.lock_idx)
                    data = self.data_from_idx(self.data_idx)
                    ccf = cross_correlate(np.array(data[1]),np.array(lock_data[1]),lags=15)
                    mydat = {0: np.array(ccf.index), 1: np.array(ccf["Cross-Correlation"])}

                    v = mydat
                    title = "Lag\t" +list(self.data.keys())[self.data_idx] + " - " + list(self.data.keys())[self.lock_idx]
                    lag_min = -25
                    lag_max = 25

                else: # auto-correlation
                    data = self.data_from_idx(self.data_idx)
                    acf = cross_correlate(np.array(data[1]),np.array(data[1]),lags=15)
                    mydat = {0: np.array(acf.index), 1: np.array(acf["Cross-Correlation"])}

                    v = mydat
                    title = "Lag\t" + list(self.data.keys())[self.data_idx] + " - " + list(self.data.keys())[self.data_idx]
                    lag_min = 0
                    lag_max = 25

                s = ""
                for i in range(2*len(data[1])-1):
                    if int(v[0][i]) >= lag_min and int(v[0][i]) <= lag_max:
                        s += str(v[0][i]) + "\t" + str(v[1][i]) + "\n"
                        print("COPY", v[0][i])

                pyperclip.copy(title + "\n" + s)
            except Exception as e:
                self.notify(str(e),title="Error")

    def update_idx(self):
        self.plot_counter = 0
        self.update()

    def update_range(self):

        try:
            if self.range_from.text() != "":
                self.data_range_min = float(self.range_from.text())
            else:
                self.data_range_min = None

            if self.range_to.text() != "":
                self.data_range_max = float(self.range_to.text())
            else:
                self.data_range_max = None
            # print("update range", self.data_range_min, self.data_range_max)
        except:
            pass

        print("UPDATE RANGE")
        self.plot_counter = 0
        self.update()

    def toggle_lock(self):

        if self.lock_slide.value() == 1:
            self.lock_idx = self.data_idx

            try:
                self.data_name_label2.setText(str(list(self.data.keys())[self.lock_idx]))
            except:
                self.data_name_label2.setText(" ")
        else:
            self.lock_idx = None
            self.data_name_label2.setText(" ")

        self.plot_counter = 0
        self.update()

    def update_idx(self):

        if self.data is not None:

            try:
                idx = int(self.table.currentRow())
                self.data_idx = idx
                self.data_name_label.setText(str(list(self.data.keys())[self.data_idx])) # "[" + str(self.data_idx)+ "] " +
                #print(self.data[list(self.data.keys())[self.data_idx]])
                # self.update()
            except Exception as e:
                self.notify("Error while updating data index\n" + str(e))


            self.plot_counter = 0
            self.update()

    def increase_idx(self):

        try:
            if self.data is None:
                return
            max_idx = len(self.data.keys())-1
            if self.data_idx < max_idx:
                self.data_idx += 1

            self.table.selectRow(self.data_idx)
            self.data_name_label.setText(str(list(self.data.keys())[self.data_idx])) # "[" + str(self.data_idx)+ "] " +

        except Exception as e:
            self.notify("Error: %s" % str(e))

        self.plot_counter = 0
        self.update()

    def decrease_idx(self):
        try:
            if self.data is None:
                return

            min_idx = 0
            if self.data_idx > min_idx:
                self.data_idx -= 1

            self.table.selectRow(self.data_idx)
            self.data_name_label.setText(str(list(self.data.keys())[self.data_idx])) # "[" + str(self.data_idx)+ "] " +


        except Exception as e:
            self.notify("Error: %s" % str(e))

        self.plot_counter = 0
        self.update()

    def load_file(self):
        fname = self.path + self.fileEdit.text()

        try:
            # load the data file

            file = open(fname,'rb')
            data = pkl.load(file)
            file.close()

            self.data = data
            # self.data_idx = 0
            # self.decrease_idx()
            # self.update()
            # self.plot_counter = 1

        except Exception as e:
            self.notify("Could not load file %s \n %s" % (fname, str(e)))


        try:
            if self.data is not None:
                self.table.clear()
                self.table.setRowCount(len(self.data.keys()))
                # insert data rows into table
                i = 0
                for k,v in self.data.items():
                    self.table.setItem(i, 0, QTableWidgetItem(str(i)))
                    self.table.setItem(i, 1, QTableWidgetItem(str(k)))

                    idx_from = int(self.idx_from_edit.text())
                    idx_to = int(self.idx_to_edit.text())

                    if isinstance(v,dict):
                        #print("IS DICT -> converting")
                        v = list(v.values())

                    if idx_to - idx_from > 1 or (idx_to == -1 and len(v) > 1 and idx_from < len(v)-1):
                        v2 = v[idx_from:idx_to]
                    else:
                        v2 = v

                    v2 = np.array(v2,dtype=np.float32)

                    #print(v2)
                    #print("--")

                    try:
                        n_vals = "%i" % (len(v2))
                        min_val = "%.3f" % (np.min(v2))
                        max_val = "%.3f" % (np.max(v2))
                        mean_val = "%.3f" % (np.mean(v2))
                        std_val = "%.3f" % (np.std(v2))
                    except:
                        n_vals = "ERR"
                        min_val = "ERR"
                        max_val = "ERR"
                        mean_val = "ERR"
                        std_val = "ERR"

                    try:
                        self.table.setItem(i, 2, QTableWidgetItem(n_vals))
                        self.table.setItem(i, 3, QTableWidgetItem(min_val))
                        self.table.setItem(i, 4, QTableWidgetItem(max_val))
                        self.table.setItem(i, 5, QTableWidgetItem(mean_val))
                        self.table.setItem(i, 6, QTableWidgetItem(std_val))
                    except:
                        pass

                    i += 1

                self.table.setHorizontalHeaderLabels(["Index","Variable","Values", "Min", "Max", "Mean", "Std"])
        except Exception as e:
            self.notify("Could not fill table: " + str(e))

        self.plot_counter = 0
        self.update()


    def data_from_idx(self,idx):
        data = [None,None]
        N = 1000
        data[0] = np.linspace(0,100,N)
        data[1] = np.full(N,42.0)+np.sin(data[0]) # *.03*np.random.rand(N) # 30*np.sin(data[0]) #

        try:
            if self.data is not None:
                current_key = list(self.data.keys())[idx]
                data[1] = self.data[current_key]
                data[0] = np.arange(len(data[1]),dtype=float)

                if isinstance(data[1],dict):
                    # print("IS DICT -> converting")
                    data[1] = list(data[1].values())

                try:
                    idx_from = int(self.idx_from_edit.text())
                    idx_to = int(self.idx_to_edit.text())

                    if idx_to - idx_from > 1 or (idx_to == -1 and len(data[0]) > 1 and idx_from < len(data[0])-1):
                        data[0] = data[0][idx_from:idx_to]
                        data[1] = data[1][idx_from:idx_to]
                except:
                    pass

        except Exception as e:

            self.notify(str(e))
        return data


    def plot_line(self,data,color=Qt.black,offset_ylabel=0.0,offset_x=0,offset_y=0,ignore_data_range=False):


        offset_x = int(offset_x)
        offset_y = int(offset_y)
        offset_ylabel= int(offset_ylabel)

        qp = QtGui.QPainter(self)
        pen = QtGui.QPen(Qt.black ,1, Qt.SolidLine)
        brush = QtGui.QBrush(Qt.white)
        qp.setBrush(brush)
        qp.setPen(pen)

        # CANVAS

        H = 550
        offset = (50,H)

        # DATA

        x = data[0]

        if len(x) < 2:
            print("X data is too short") # self.notify("X data is too short")
            return

        x_max_label = "%i"% int(np.nanmax(x))
        x_min_label = "%i"% int(np.nanmin(x))
        x_mid_label = "%i"% int(0.5*(np.nanmax(x)+ np.nanmin(x)))


        x = x-np.nanmin(x)
        if np.nanmax(x) != 0.0:
            x /=np.nanmax(x)
        x*= 200

        scale = 3
        dx = 20

        x_max = int(np.nanmax(x))
        x_min = int(np.nanmin(x))
        x_mid = int(0.5*(np.nanmin(x) + np.nanmax(x)))

        y = data[1]

        y_max_label = "%.2f"% np.nanmax(y)
        y_min_label = "%.2f"% np.nanmin(y)
        y_mid_label = "%.2f"% (0.5*(np.nanmax(y)+ np.nanmin(y)))

        if not ignore_data_range:

            if self.data_range_min is not None:
                min_yrange = self.data_range_min
            else:
                min_yrange = np.nanmin(y)

            if self.data_range_max is not None:
                max_yrange = self.data_range_max
            else:
                max_yrange = np.nanmax(y)
        else:
            min_yrange= np.nanmin(y)
            max_yrange = np.nanmax(y)


        y = np.array(y)-min_yrange

        norm = (max_yrange-min_yrange)
        if norm != 0.0:
            y /= norm


        y*= 200
        y +=  int(offset_y)

        if np.nanmin(y) == np.nanmax(y):
            y += 100 # shift to center
            y_max_label = ""
            y_min_label = ""

        # DRAW AXIS
        y_max = int(np.nanmax(y))
        y_min = int(np.nanmin(y))
        y_mid = int(0.5*(np.nanmin(y) + np.nanmax(y)))

        if y_max == y_min:
            y_max = y_mid + 100

        points_yticks = [(x_max,y_min), (x_max,y_mid), (x_max,y_max)]
        points_xticks = [(x_min,0), (x_mid,0), (x_max,0)]

        pen = QtGui.QPen(Qt.gray ,1, Qt.SolidLine)
        qp.setPen(pen)

        for p in points_yticks:
            u = p[0]*scale
            v = p[1]
            qp.drawLine(QtCore.QPoint(int(offset[0]+u+dx),int(H-v)),QtCore.QPoint(int(offset[0]+u+10+dx),int(H-v)))

        for p in points_xticks:
            u = int(p[0]*scale +dx )
            v = p[1]- 10
            qp.drawLine(QtCore.QPoint(int(offset[0]+u),int(H-v-10)),QtCore.QPoint(int(offset[0]+u),int(H-v)))

        r1 = QtCore.QPoint(int(offset[0]+0+dx),int(H-offset_y))
        r2 = QtCore.QPoint(int(offset[0]+x_max*scale +dx) ,int(H-offset_y))
        r3 = QtCore.QPoint(int(offset[0]+x_max*scale +dx),int(H-(200)-offset_y))
        r4 = QtCore.QPoint(int(offset[0]+0+dx),int(H-(200)-offset_y))
        qp.drawLine(r1,r2)
        qp.drawLine(r2,r3)
        qp.drawLine(r3,r4)
        qp.drawLine(r4,r1)

        lx1 = QtCore.QPoint(int(offset[0]+dx+points_xticks[0][0]*scale),int(H-points_xticks[0][1]+25)-offset_y)
        lx2 = QtCore.QPoint(int(offset[0]+dx+points_xticks[1][0]*scale),int(H-points_xticks[1][1]+25)-offset_y)
        lx3 = QtCore.QPoint(int(offset[0]+dx+points_xticks[2][0]*scale),int(H-points_xticks[2][1]+25)-offset_y)

        ly1 = QtCore.QPoint(int(offset[0]+dx+points_yticks[0][0]*scale)+25+offset_ylabel,int(H-points_yticks[0][1])+5)
        ly2 = QtCore.QPoint(int(offset[0]+dx+points_yticks[1][0]*scale)+25+offset_ylabel,int(H-points_yticks[1][1])+5)
        ly3 = QtCore.QPoint(int(offset[0]+dx+points_yticks[2][0]*scale)+25+offset_ylabel,int(H-points_yticks[2][1])+5)

        qp.drawText(lx1,x_min_label)
        qp.drawText(lx2,x_mid_label)
        qp.drawText(lx3,x_max_label)

        pen = QtGui.QPen(color ,1, Qt.SolidLine)
        qp.setPen(pen)

        qp.drawText(ly1,y_min_label)
        qp.drawText(ly2,y_mid_label)
        qp.drawText(ly3,y_max_label)

        # DRAW DATA

        # pen = QtGui.QPen(color ,1.0, Qt.SolidLine)
        # qp.setPen(pen)

        qp.drawPoint(QtCore.QPoint(int(offset[0]+x[0]),int(H-y[0])))

        for i in range(1,len(x)):

            # qp.drawPoint(QtCore.QPoint(offset[0]+xi,(H-yi)))
            cond1 = not (np.isnan(x[i-1]) or np.isnan(x[i]) or np.isnan(y[i-1]) or np.isnan(y[i]))

            x1_transformed = int(dx + offset[0]+scale*x[i-1])
            y1_transformed = int(H-y[i-1])
            x2_transformed = int(dx + offset[0]+scale*x[i])
            y2_transformed = int(H-y[i])

            cond2 = (offset_y <= y[i-1] <= 200 + offset_y ) and (offset_y <= y[i] <= 200+offset_y)

            if cond1 and cond2:

                pstart = QtCore.QPoint(x1_transformed,y1_transformed)
                pend = QtCore.QPoint(x2_transformed,y2_transformed)

                qp.drawLine(pstart, pend)


    def paintEvent(self,event):

        update_data = False  # only update every N time steps
        if self.plot_counter <= 0: # or len(self.plot_data) == 0:
            self.plot_counter = self.plot_rep
            update_data = True

        self.plot_counter -= 1

        if update_data:
            data = self.data_from_idx(self.data_idx)
            self.plot_data["data"] = data
        else:
            data = self.plot_data["data"] # read from memory instead of computing again


        try:
            self.plot_line(data)

        except Exception as e:
            print("Error:", str(e))
            #self.notify(str(e))

        # OVERLAY LINE IN RED
        if self.lock_idx is not None:
            if update_data:
                lock_data = self.data_from_idx(self.lock_idx)
                self.plot_data["lock_data"] = lock_data
            else:
                lock_data = self.plot_data["lock_data"]

            try:
                self.plot_line(lock_data,color=Qt.red,offset_ylabel=50)
            except Exception as e :
                print("error",str(e))
                #self.notify(str(e))


        # CROSS-CORRELATION
        try:
            if self.lock_idx is not None: # plot cross-correlation
                if update_data:
                    ccf = cross_correlate(np.array(data[1]),np.array(lock_data[1]),lags=15)
                    mydat = {0: ccf.index, 1: ccf["Cross-Correlation"]}
                    self.plot_data["mydat"] = mydat
                else:
                    mydat = self.plot_data["mydat"]

                self.plot_line(mydat,color=Qt.blue,offset_y = -300,ignore_data_range=True)

            else: # plot auto-correlation
                if update_data:
                    acf = cross_correlate(np.array(data[1]),np.array(data[1]),lags=15)
                    mydat = {0: acf.index, 1: acf["Cross-Correlation"]}
                    self.plot_data["mydat"] = mydat
                else:
                    mydat = self.plot_data["mydat"]
                self.plot_line(mydat,color=Qt.blue,offset_y = -300,ignore_data_range=True)



        except Exception as e:
            print("Error:",str(e))
            # self.notify(str(e))

        # print("COUNT", self.plot_counter)



if __name__ == "__main__":
    """
    """

    app = QtWidgets.QApplication(sys.argv)
    # window = ProjectDesigner()
    window = OutputDisplay(None)
    app.exec_()



from PyQt5 import QtWidgets, uic,QtGui
from PyQt5.QtCore import QUrl
from PyQt5.QtWidgets import QFileDialog, QMessageBox, QTableWidgetItem, QAbstractItemView, QListWidgetItem, QShortcut
import sys
import os
import yaml
import shutil

# from pandasmodel import PandasModel
from PyQt5.QtGui import QDesktopServices
import pickle as pkl
import re
import matplotlib.pyplot as plt
from PyQt5.QtCore import Qt
from PyQt5 import QtCore
from PyQt5.QtGui import QKeySequence

from .mamba_interpreter2 import convert_code

from PyQt5.QtGui import QFont, QFontDatabase, QColor, QSyntaxHighlighter, QTextCharFormat
from PyQt5.QtGui import QKeySequence, QFontMetrics
from PyQt5.QtCore import QObject, QThread, pyqtSignal
import time


class WatchDog(QObject):
    """
    This is a file watchdog. It frequently checks if external changes
    have been made to a file viewed in the code editor.
    """
    difference_detected = pyqtSignal()  # difference detected between current code and code stored in file

    def __init__(self,agent_name,file_name,parent):

        super().__init__()
        self.agent_name = agent_name
        #self.file_name = os.path.join(file_name , agent_name + ".txt" ) # < source file sub_names
        self.file_name = os.path.join(file_name , agent_name + ".py" ) # < source file sub_names
        self.parent_w = parent
        text = parent.textEdit.toPlainText()
        self.finished = False


    def run(self):
        """Long-running task."""
        while not self.finished:
            time.sleep(1.0) # check every second

            # print("check file", self.agent_name,self.file_name)
            current_gui_text = self.parent_w.textEdit.toPlainText()
            current_file_text = None

            try:
                with open(self.file_name,"r") as file:
                    current_file_text = file.read()

                outdated = False
                file_time = os.path.getmtime(self.file_name)
                my_time = self.parent_w.mtime

                if self.parent_w.agent_name not in CodeEditor.instances:
                    self.finished = True

                if file_time > my_time:
                    outdated = True

                if current_gui_text != current_file_text and outdated:
                    self.difference_detected.emit()

            except Exception as e:
                print(str(e))
                # pass


class Highlighter(QSyntaxHighlighter):
    def __init__(self,parent=None):
        super().__init__(parent)

        # self._mapping = {}
        self._rules = []

        fmt_comment = QTextCharFormat()
        # fmt_comment.setFontItalic(True)
        fmt_comment.setForeground(QtGui.QColor("#239e77"))

        self.tri_single = (QtCore.QRegExp("'''"), 1, fmt_comment)
        self.tri_double = (QtCore.QRegExp('"""'), 2, fmt_comment)

    def build_rules(self):
        # Build a QRegExp for each pattern
        my_rules = []
        for r in self._rules:
            try:
                pat, index, fmt = r
                try:
                    new_pat = QtCore.QRegExp(pat)
                    my_rules.append((new_pat, index, fmt))
                except Exception as e:
                    print("EXCEPTION IN REGEXP", str(e))

            except Exception as e:
                print("EXCEPTION", str(e))
        self._rules = my_rules

    def add_mapping(self, pattern, pattern_format):
        # self._mapping[pattern] = pattern_format
        pattern = str(pattern).replace("$","\$").replace("*","\*").replace("(", "\(").replace(")", "\)")

        # new_mapping = "^" + pattern + "$"
        self._rules.append((pattern, 0, pattern_format))

    def match_multiline(self, text,delimiter,in_state,style):

            if self.previousBlockState() == in_state:
                start = 0
                add = 0

            else:
                start = delimiter.indexIn(text)
                if start in self.tripleQuoutesWithinStrings:
                    return False

                add = delimiter.matchedLength()

            while start >= 0:
                end = delimiter.indexIn(text,start+add)

                if end >= add:
                    length = end-start + add + delimiter.matchedLength()
                    self.setCurrentBlockState(0)

                else:
                    self.setCurrentBlockState(in_state)
                    length = len(text) - start + add

                self.setFormat(start,length,style)
                start = delimiter.indexIn(text,start+length)

            if self.currentBlockState() == in_state:
                return True
            else:
                return False

    def highlightBlock(self, text_block):
        # print("highligth block", text_block, )

        try:
            self.tripleQuoutesWithinStrings = []

            nth = 0
            for expression, nth, fmt in self._rules: # self._mapping.items():

                index = expression.indexIn(text_block, 0)
                if index >= 0:
                    if expression.pattern() in [r'"[^"\\]*(\\.[^"\\]*)*"', r"'[^'\\]*(\\.[^'\\]*)*'"]:
                        innerIndex = self.tri_single[0].indexIn(text_block, index + 1)
                        if innerIndex == -1:
                            innerIndex = self.tri_double[0].indexIn(text_block, index + 1)
                        if innerIndex != -1:
                            tripleQuoteIndexes = range(innerIndex, innerIndex + 3)
                            self.tripleQuoutesWithinStrings.extend(tripleQuoteIndexes)


                while index>=0:
                    # skip triple quotes
                    if index in self.tripleQuoutesWithinStrings:
                        index += 1
                        expression.indexIn(text_block,index)
                        continue

                    index = expression.pos(nth)
                    length = len(expression.cap(nth))
                    self.setFormat(index,length,fmt)
                    index = expression.indexIn(text_block,index+length)

            self.setCurrentBlockState(0)

            # do multi-line strings
            in_multiline = self.match_multiline(text_block,*self.tri_single)
            if not in_multiline:
                in_multiline = self.match_multiline(text_block,*self.tri_double)


        except Exception as e:
            print(str(e))
        




class CodeEditor(QtWidgets.QWidget):

    instances = {}

    def closeEvent(self, event):
        del self.__class__.instances[self.agent_name]
        event.accept() # let the window close


    def paintEvent(self,event):
        pass

    def __init__(self,parent,box,text=""):

        super(CodeEditor, self).__init__(parent) # Call the inherited classes __init__ method

        self.__class__.instances[box.name] = self
        path = os.path.dirname(os.path.abspath(__file__))
        uic.loadUi(os.path.join(path,'code_editor.ui'), self) # Load the .ui file

        project_file = box.parent_widget.main_widget.current_file
        try:
            self.mtime = os.path.getmtime(project_file) # last changes of project file
        except Exception as e:
            print(str(e))


        self.move(2,2)
        self.setGeometry(2,2,parent.frameGeometry().width()-3,parent.frameGeometry().height()-3)
        self.backup_count = 0

        self.main_widget = box.parent_widget.main_widget

        theme_manager = box.parent_widget.main_widget.theme_manager
        self.setStyleSheet(theme_manager.get_stylesheet("main"))
        self.setStyleSheet(theme_manager.get_background_style())

        self.large_size = QtCore.QSize(1569,874)
        self.small_size = QtCore.QSize(950,874)
        self.size = "small"
        #self.pushButton.pressed.connect(self.switch_size)

        self.search_idx = -1
        self.matched_lines = []
        self.matched_lines_idx = 0

        #self.shortcut = QShortcut(QKeySequence("Ctrl+S"),self)
        #self.shortcut.activated.connect(self.save_and_build)
        self.keylist = []

        # self.shortcut.activated.connect(lambda: print("ACTIVATED"))

        self.setStyleSheet(theme_manager.get_stylesheet("main"))
        self.setStyleSheet(theme_manager.get_background_style())

        self.textEdit.installEventFilter(self)
        self.textEdit.cursorPositionChanged.connect(self.update_line_label)

        self.linedict = {}

        self.shortcut3 = QShortcut(QKeySequence("Shift+Tab"),self)
        self.shortcut3.activated.connect(self.unindent)

        self.shortcut4 = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Plus),self)
        self.shortcut4.activated.connect(self.zoom_in)

        self.shortcut5 = QShortcut(QKeySequence(Qt.CTRL +Qt.Key_Minus),self)
        self.shortcut5.activated.connect(self.zoom_out)

        # self.setFixedSize(self.small_size)
        # self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # self.setSizeGripEnabled(True)
        self.highlighter = Highlighter(self)
        self.setUpEditor()

        if text == "":
            text = "$[AGENT] %s\n+[INIT]\npass # < do nothing \n+[ENDINIT]\n\n$[END]\n" % (box.name)

        self.textEdit.setPlainText(text)
        self.textEdit.textChanged.connect(self.update_interpreter)
        self.searchEdit.textChanged.connect(self.update_search)

        self.box = box
        self.agent_name = self.box.name
        try:
            self.setWindowTitle("Edit Agent " + str(self.box.name).capitalize())
        except Exception as e:
            print(str(e))


        self.highlighter_rects = [] # highlighter for functions
        self.highlighter_tans  = [] # highlighter for transactions
        self.highlighter_bal   = [] # highlighter for balance sheet manipulations

        self.count_interpr_max = 5
        self.count_update_max = 5
        self.count_interpr = self.count_interpr_max
        self.count_update = self.count_update_max
        # self.okButton.pressed.connect(self.send_data)

        self.nextButton.pressed.connect(self.update_matched_line)
        self.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,QtWidgets.QSizePolicy.Expanding))
        self.textEdit2.setVisible(False)

        # self.main_widget.resizeEvent(None)
        parent.resizeEvent = self.resizeEvent
        parent.parent().resizeEvent = self.resizeEvent
        parent.parent().parent().resizeEvent = self.resizeEvent

        self.show()

        # file watchdog
        self.thread = QThread()
        filepath = os.path.dirname(os.path.abspath(self.main_widget.current_file))
        mamba_path = os.path.join(filepath , "mamba_code")
        self.worker = WatchDog(self.agent_name,mamba_path,self)
        self.worker.moveToThread(self.thread)
        self.worker.difference_detected.connect(self.process_differences)
        self.thread.started.connect(self.worker.run)
        self.thread.start()

        self.tab = None

        self.closeButton.pressed.connect(self.close_editor)


    def zoom_in(self):
        self.textEdit.zoomIn()
        self.update()

    def zoom_out(self):
        self.textEdit.zoomOut()
        self.update()

    def close_editor(self):
        self.worker.finished = True

        try:
            tw = self.main_widget.tabWidget
            tw.removeTab(tw.indexOf(self.tab))

            self.__class__.instances[self.agent_name] = None
            del self.__class__.instances[self.agent_name]

        except Exception as e:
            print(str(e))

    def process_differences(self):
        # process differences between the sfctl file and the external file changes
        self.warn_differences()
        self.set_text_from_external()

    def set_text_from_external(self):
        """
        set the text from the external file changes. Overwrite internal text
        """

        filepath = os.path.dirname(os.path.abspath(self.main_widget.current_file))
        mamba_path = os.path.join(filepath , "mamba_code")

        new_text = None
        with open(os.path.join(mamba_path , self.agent_name + ".py"),"r") as file:
            new_text = file.read()

        if new_text is None:
            return

        scroll_pos = self.textEdit.verticalScrollBar().value()

        cursor = self.textEdit.textCursor()
        old_position= cursor.position()

        self.textEdit.setPlainText(new_text)

        cursor = self.textEdit.textCursor()
        cursor.setPosition(old_position)

        self.textEdit.setTextCursor(cursor)
        self.textEdit.verticalScrollBar().setValue(scroll_pos)


    def warn_differences(self):
        # self.main_widget.notify("Difference Detected!\n%s" % self.agent_name,title="External changes detected.")
        self.main_widget.statusBar().showMessage("External changes in %s detected! Overwriting..." % (self.agent_name))

    def resizeEvent(self, event):
        # print("CODE EDITOR RESIZE")
        self.resize_frame()

    def resize_frame(self):
        parent = self.parent()
        self.setGeometry(2,2,parent.frameGeometry().width()-2,parent.frameGeometry().height()-2)

    def search_name(self):
        # search for name of the edited agent
        text = self.textEdit.toPlainText()
        new_name = "None"
        for line in text.split("\n"):
            for searchstr in ["$[AGENT]", "$[CLASS]" , "$[MARKET]"]:
                if line.startswith(searchstr):

                    mydef = line.split(searchstr)[1].strip().split("(")[0].strip()
                    new_name = mydef[0] + mydef[1:].split(".")[0]

                    return new_name

        return new_name # should not happen or else the file is 'bad'


    def save_and_build(self):
        """
        updates the files and re-builds project
        """
        print("save and build")


        # check if the agent was renamed (?)
        original_name = self.box.name
        new_name = self.search_name()

        # self.main_widget.notify("Checking name: %s, %s" %(original_name,new_name),title="name checking" )

        if original_name != new_name:

            if new_name not in self.main_widget.code_data:

                self.worker.agent_name = new_name
                # check if agent name already exists. if not, migrate code to new name

                old_data = self.textEdit.toPlainText() # self.main_widget.code_data[original_name]
                self.main_widget.code_data[new_name] = old_data

                # if original_name in self.main_widget.code_data:
                #   del self.main_widget.code_data[original_name]

                self.box.name = new_name
                self.main_widget.notify("The name of your agent has changed from %s to %s." % (original_name,new_name), title="Info")

                tabwidget = self.parent().parent().parent()
                tab = self.tab # parent().parent()

                if tab is not None:
                    tab_i = tabwidget.indexOf(tab)
                    #tabwidget.removeTab(tab_i)
                    #tabwidget.setTabVisible(tab_i,False)
                    tabwidget.setTabText(tab_i,new_name)

                self.main_widget.update_table()


            else:
                self.main_widget.notify("An agent with the name %s already exists. Keeping %s." % (new_name,original_name), title="Error")


            return original_name,new_name,self

        return None

        #
        # self.main_widget.save_and_build()

    def update_matched_line(self):

        # update matched lines
        try:
            searchstr = self.searchEdit.text()
            print("SEARCHSTR*", searchstr)
            mystring = self.textEdit.toPlainText()
            self.matched_lines = []
            for i,line in enumerate(mystring.split('\n')):
                if searchstr in line:
                    self.matched_lines.append(i)

            # update scrolling

            if len(self.matched_lines) <= 0:
                return

            self.matched_lines_idx += 1
            if self.matched_lines_idx >= len(self.matched_lines):
                self.matched_lines_idx = 0
            # print("matched_lines",self.matched_lines)

            # print("IDX",self.matched_lines_idx)
            matchline = self.matched_lines[self.matched_lines_idx]
            self.textEdit.verticalScrollBar().setValue(max(0,matchline - 1))
        except Exception as e:
            print(str(e))
        
        self.update_line_highlight()


    def paintEvent(self,event):

        qp = QtGui.QPainter(self)
        pen = QtGui.QPen(Qt.black ,1, Qt.SolidLine)
        brush = QtGui.QBrush(Qt.black)
        qp.setBrush(brush)
        qp.setPen(pen)

        font = self.textEdit.document().defaultFont()
        metrics = QFontMetrics(font)

        y_rect_os = 48

        for block_number in self.highlighter_rects:
            offset = self.textEdit.contentOffset()

            block= self.textEdit.document().findBlockByNumber(block_number)
            brect = self.textEdit.blockBoundingGeometry(block)
            # print("brect", brect)
            y = int(brect.y()+ offset.y()) + y_rect_os # 53 #  53
            x = int(brect.x() + 6)
            w = brect.width()
            h = brect.height()

            if(y >= y_rect_os and y <= 920):
                qp.drawRect(x+38,y+4,7,7)

        pen = QtGui.QPen(QtGui.QColor("#a69819") ,1, Qt.SolidLine)
        qp.setPen(pen)
        brush = QtGui.QBrush(QtGui.QColor("#a69819"))
        qp.setBrush(brush)

        for block_number in self.highlighter_tans:
            offset = self.textEdit.contentOffset()

            block= self.textEdit.document().findBlockByNumber(block_number)
            brect = self.textEdit.blockBoundingGeometry(block)
            # print("brect", brect)
            y = int(brect.y()+ offset.y()) + y_rect_os #  53
            x = int(brect.x() + 6)
            w = brect.width()
            h = brect.height()

            if(y >= y_rect_os and y <= 920):
                qp.drawRect(x+53,y+4,7,7)

        pen = QtGui.QPen(QtGui.QColor("#4bd16f") ,1, Qt.SolidLine)
        qp.setPen(pen)
        brush = QtGui.QBrush(QtGui.QColor("#4bd16f"))
        qp.setBrush(brush)

        for block_number in self.highlighter_bal:
            offset = self.textEdit.contentOffset()

            block= self.textEdit.document().findBlockByNumber(block_number)
            brect = self.textEdit.blockBoundingGeometry(block)
            # print("brect", brect)
            y = int(brect.y()+ offset.y()) + y_rect_os #  53
            x = int(brect.x() + 6)
            w = brect.width()
            h = brect.height()

            if(y >= y_rect_os and y <= 920):
                qp.drawRect(x+53,y+4,7,7)


        pen = QtGui.QPen(Qt.gray ,1, Qt.SolidLine)
        qp.setPen(pen)
        brush = QtGui.QBrush(Qt.gray)
        qp.setBrush(brush)

        for block_number in range(self.textEdit.document().lineCount()):

            offset = self.textEdit.contentOffset()

            block= self.textEdit.document().findBlockByNumber(block_number)
            brect = self.textEdit.blockBoundingGeometry(block)
            # print("brect", brect)
            y = int(brect.y()+ offset.y()) + 53 #  53
            x = int(brect.x() + 6)
            w = brect.width()
            h = brect.height()

            if(y >= 50 and y <= 920) and int(block_number)%5 == 0:
                try:

                    qp.drawText(x,y+9,"%04i" % (self.linedict[int(block_number)]+1))
                except:
                    self.update_interpreter()
                    try:
                        qp.drawText(x,y+9,"--")
                    except:
                        pass

    def keyPressEvent(self,event):
        self.keylist.append(event.key())
        # pass
        # ctrl + s is handled in main widget

    def keyReleaseEvent(self,event):
        self.keylist = []

    def eventFilter(self, obj, event):
        # print("event",event)

        if event.type() == QtCore.QEvent.KeyPress and obj is self.textEdit:

            self.mtime = time.time()

            if event.key() == QtCore.Qt.Key_Tab: # and self.text_box.hasFocus():
                # print('Tab pressed')
                self.indent()
                return True

            elif event.key() == QtCore.Qt.Key_Return:
                self.new_line()
                return True

        elif event.type() == QtCore.QEvent.Wheel:

            self.update_line_highlight()
            self.count_interpr = 0
            self.update_interpreter()
            self.repaint()

        self.count_update -=1
        if self.count_update <= 0:
            self.update()
            self.count_update = self.count_update_max

        return False

    def new_line(self):

        # insert new line with previous indentation
        scroll_pos = self.textEdit.verticalScrollBar().value()
        cursor = self.textEdit.textCursor()
        old_position= cursor.position()
        text = self.textEdit.toPlainText()

        try:
            my_line = text.splitlines()[cursor.blockNumber()]
        except:
            my_line = text.splitlines()[cursor.blockNumber()-1] # if is last line
        indentation = max(0,len(my_line.split("\t"))-1)

        if my_line.strip().endswith(":"):
            indentation += 1

        new_str = "\n" + "\t"*indentation
        self.textEdit.setPlainText(text[:old_position] + new_str + text[old_position:])
        cursor.setPosition(old_position + len(new_str))

        self.textEdit.setTextCursor(cursor)
        self.textEdit.verticalScrollBar().setValue(scroll_pos)

    def indent(self):
        # inert tab for whole selection

        # return # NOTE deactivate due to performance issues?
        # TODO fix performance, possibly due to for loop when dealing with large files ?

        scroll_pos = self.textEdit.verticalScrollBar().value()

        cursor = self.textEdit.textCursor()
        old_position= cursor.position()

        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        text = self.textEdit.toPlainText()

        mystr = text[:start]
        my_lines = []

        for line in text[start:end].split("\n"):
            my_lines.append("\t" + line)

        new_str =  "\n".join(my_lines)
        new_end = start + len(new_str)
        mystr += new_str

        mystr += text[end:]

        self.textEdit.setPlainText(mystr)

        cursor = self.textEdit.textCursor()
        if new_end - start > 1:

            cursor.setPosition(start)
            cursor.setPosition(new_end, QtGui.QTextCursor.KeepAnchor)
        else:
            cursor.setPosition(new_end)

        self.textEdit.setTextCursor(cursor)
        self.textEdit.verticalScrollBar().setValue(scroll_pos)

    def unindent(self):
        # inert tab for whole selection

        scroll_pos = self.textEdit.verticalScrollBar().value()

        cursor = self.textEdit.textCursor()
        old_position= cursor.position()

        start = cursor.selectionStart()
        end = cursor.selectionEnd()

        text = self.textEdit.toPlainText()

        mystr = text[:start]
        my_lines = []

        for line in text[start:end].split("\n"):
            if line.startswith("\t"):
                my_lines.append(line[1:])
            else:
                my_lines.append(line)

        new_str =  "\n".join(my_lines)
        new_end = start + len(new_str)
        mystr += new_str

        mystr += text[end:]

        self.textEdit.setPlainText(mystr)

        cursor = self.textEdit.textCursor()

        if new_end - start > 1:

            cursor.setPosition(start)
            cursor.setPosition(new_end, QtGui.QTextCursor.KeepAnchor)
        else:
            cursor.setPosition(new_end)

        self.textEdit.setTextCursor(cursor)
        self.textEdit.verticalScrollBar().setValue(scroll_pos)


    def switch_size(self):

        if self.size == "large":
            self.size = "small"
            self.setFixedSize(self.small_size)

        elif self.size == "small":
            self.size = "large"
            self.setFixedSize(self.large_size)

        self.count_interpr = 0
        self.update_interpreter()


    def update_interpreter(self):
        """
        interpret each line of code here
        """

        # old_text = self.textEdit.toPlainText()
        # self.textEdit.setPlainText(str(old_text.encode("utf-8").decode('cp1252')))

        scroll_pos = self.textEdit2.verticalScrollBar().value()

        if self.count_interpr <= 0 or self.size=="large":
            self.count_interpr = self.count_interpr_max

            code_lines = self.textEdit.toPlainText().split("\n")
            translated_code,mydict = convert_code(code_lines)
            self.textEdit2.setPlainText(translated_code)
            self.linedict = mydict

        self.count_interpr -= 1

        self.textEdit2.verticalScrollBar().setValue(scroll_pos)

        self.update_line_highlight()
        self.send_data()

    def update_line_highlight(self):
        code_lines = self.textEdit.toPlainText().split("\n")

        self.highlighter_rects = []
        self.highlighter_tans = []
        self.highlighter_bal = []

        for i,line in enumerate(code_lines):
            if line.startswith("+[FUN]") or line.startswith("+[INIT]") or line.startswith("+[PARAM]") or line.startswith("+[KNOWS]"):
                self.highlighter_rects.append(i)
            elif "<~~>" in line:
                self.highlighter_tans.append(i)

            elif "ASSETS" in line or "LIABILITIES" in line or "EQUITY" in line:
                self.highlighter_bal.append(i)

        # print("highlighter_rects", self.highlighter_rects)

    def update_line_label(self):
        cursor = self.textEdit.textCursor()
        x = cursor.blockNumber() + 1
        y = cursor.columnNumber() + 1
        z = -1

        # print("linedict",self.linedict)

        if int(x) in self.linedict:
            z = int(self.linedict[int(x)]) - 2
            # print("z = ", z)

        if z < 0:
            z = -1

        #self.lineLabel.setText('Row {0:d}, Col {1:d} Python {2:d}'.format(x,y,z))
        self.lineLabel.setText('Line {0:d}, Python: {1:d}'.format(x,z))

    def update_search(self):
        self.setUpEditor()
        self.update_interpreter()

    def send_data(self):
        self.box.parent_widget.code_data[self.box.name] = str(self.textEdit.toPlainText())
        # self.close()
        if self.backup_count == 1000:
            self.main_widget.auto_backup()
            self.backup_count = 0
        self.backup_count += 1

    def setUpEditor(self):

        self.highlighter = Highlighter(self)

        # Agent
        agent_format = QTextCharFormat()
        #agent_format.setFontWeight(QFont.Bold)
        agent_format.setForeground(QtGui.QColor("#ad3836"))
        pattern = r'\$\[[A-Za-z0-9]+\]'
        self.highlighter.add_mapping(pattern,agent_format)

        # pythonic syntax
        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        fmt.setFontWeight(QFont.Bold)
        # fmt.setForeground(QtGui.QColor("#d42408"))
        for i in ["self","if ","elif ", "else", "for ","while ", " in ", " not ", "True", "False","and ", "or "]: #, "\(", "\)"]:
            self.highlighter.add_mapping(i,fmt)
            # TODO hide this in comments

        fmt = QTextCharFormat()
        #fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#636363"))
        for i in ["\(","\)"]:
            self.highlighter.add_mapping(i,fmt)


        import_format = QTextCharFormat()
        #import_format.setFontWeight(QFont.Bold)
        import_format.setForeground(QtGui.QColor("#635b5a"))
        pattern = r'\+\[IMPORT\]'
        self.highlighter.add_mapping(pattern,import_format)

        fmt = QTextCharFormat()
        #fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#e35a44"))
        pattern = r'\+\[KNOWS\]'

        self.highlighter.add_mapping(pattern,fmt)


        fmt = QTextCharFormat()
        #fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#e35a44"))
        pattern = r'\+\[PARAM\]'
        self.highlighter.add_mapping(pattern,fmt)

        fmt = QTextCharFormat()
        # fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#cf2115"))
        pattern = r'\+\[FUN\]'
        self.highlighter.add_mapping(pattern,fmt)

        fmt = QTextCharFormat()
        # fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#cf2115"))
        pattern = r'\+\[ENDFUN\]'
        self.highlighter.add_mapping(pattern,fmt)


        fmt = QTextCharFormat()
        #fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#ff6d63"))
        pattern = r'\$\.[A-Za-z0-9_]+'
        self.highlighter.add_mapping(pattern,fmt)


        fmt = QTextCharFormat()
        # mt.setFontWeight(QFont.It)
        fmt.setForeground(QtGui.QColor("#9c8681"))
        pattern = r'\<\~\~\>\s+[A-Za-z0-9,\@\s\_]+'
        self.highlighter.add_mapping(pattern,fmt)

        fmt = QTextCharFormat()
        fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#80473b"))
        pattern = r'\@[A-Za-z0-9_]+'
        self.highlighter.add_mapping(pattern,fmt)

        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#3d83a8"))
        self.highlighter.add_mapping(r'balance\_sheet',fmt)
        self.highlighter.add_mapping(r'income\_statement',fmt)

        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#614d70"))
        self.highlighter.add_mapping(r'update',fmt)
        self.highlighter.add_mapping(r'file_bankruptcy',fmt)

        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        # fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#d42408"))
        self.highlighter.add_mapping(r'\+\[INIT\]',fmt)
        self.highlighter.add_mapping(r'\+\[ENDINIT\]',fmt)

        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        # fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#366630"))
        self.highlighter.add_mapping(r'\+\[ACCOUNTING\]',fmt)
        self.highlighter.add_mapping(r'\+\[ENDACCOUNTING\]',fmt)


        fmt = QTextCharFormat()
        fmt.setForeground(QtGui.QColor("#60add6"))
        for element in ["gross_income","gross_spendings","ebit","net_income","noi","get_entry","get_balance","net_worth",
                        "total_assets","total_liabilities"]:
        #["REVENUES","GAINS","EXPENSES","LOSSES", "NONTAX_PROFITS", "NONTAX_LOSSES","INTEREST", "TAXES", "NOI",

            self.highlighter.add_mapping(element,fmt)

        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        # fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#4bd16f"))
        for i in ["ASSETS", "EQUITY", "LIABILITIES"]+["REVENUES", "GAINS", "EXPENSES", "LOSSES", "NONTAX_PROFITS", "NONTAX_LOSSES", "INTEREST", "TAXES", "NOI"]:
            self.highlighter.add_mapping(i,fmt)

        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        # fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#4bd16f"))
        self.highlighter.add_mapping("BALANCE?.",fmt)
        self.highlighter.add_mapping("INCOME?.",fmt)

        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        # fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#888f8a"))
        self.highlighter.add_mapping("<>",fmt)
        self.highlighter.add_mapping("<.>",fmt)


        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        # fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#bd1a91"))
        self.highlighter.add_mapping("\$URAND",fmt)

        
        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        # fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#bd1a91"))
        self.highlighter.add_mapping("\$NRAND",fmt) # <- TODO add to mamba interpreter


        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        # fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#b3a834"))
        self.highlighter.add_mapping("\!LOG",fmt)  # <- TODO add to mamba interpreter

        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#334f7a"))
        self.highlighter.add_mapping("print",fmt)
        
        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        fmt.setFontWeight(QFont.Bold)
        #fmt.setForeground(QtGui.QColor("#bd1a91"))
        self.highlighter.add_mapping("return",fmt)
        self.highlighter.add_mapping("pass",fmt)
        self.highlighter.add_mapping("rematch",fmt)

        fmt = QTextCharFormat()
        # fmt.setFontWeight(QFont.Bold)
        fmt.setForeground(QtGui.QColor("#ababab"))
        pattern = r'\#.+'
        self.highlighter.add_mapping(pattern,fmt)

        searchtext = self.searchEdit.text()
        
        if searchtext != "":
            
            try:
                # searchtext = searchtext.replace("$", "").replace("*", "")
                fmt = QTextCharFormat()
                # fmt.setFontItalic(True)
                fmt.setFontWeight(QFont.Bold)
                fmt.setForeground(QtGui.QColor("#ffff2e"))
                fmt.setBackground(QtGui.QColor("#000000"))
                self.highlighter.add_mapping(searchtext, fmt)
                
            except Exception as e:
                print(str(e))

        # self.textEdit.setPlainText(self.textEdit.toPlainText())

        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        fmt.setForeground(QtGui.QColor("#5a14ff"))
        fmt.setBackground(QtGui.QColor("#dbdbdb"))
        self.highlighter.add_mapping("NOTE",fmt)


        fmt = QTextCharFormat()
        # fmt.setFontItalic(True)
        fmt.setForeground(QtGui.QColor("#ff4f14"))
        fmt.setBackground(QtGui.QColor("#dbdbdb"))
        self.highlighter.add_mapping("TODO",fmt)

        self.highlighter.build_rules()

        self.highlighter.setDocument(self.textEdit.document())




if __name__ == "__main__":
    """
    """

    app = QtWidgets.QApplication(sys.argv)
    # window = ProjectDesigner()
    window = CodeEditor()
    app.exec_()



import sys
import pandas as pd
from PyQt5.QtWidgets import QApplication, QTableView
from PyQt5.QtCore import QAbstractTableModel, Qt
from PyQt5 import QtCore


class PandasModel(QAbstractTableModel):

    def __init__(self, data):
        QAbstractTableModel.__init__(self)
        self._data = data

    def rowCount(self, parent=None):
        return self._data.shape[0]

    def columnCount(self, parnet=None):
        return self._data.shape[1]

    def get_rawdata(self):
        return self._data.copy()

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():
            if role == Qt.DisplayRole:
                return str(self._data.iloc[index.row(), index.column()])
        return None

    def headerData(self, x, orientation, role):
        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return self._data.columns[x]
        if orientation == QtCore.Qt.Vertical and role == QtCore.Qt.DisplayRole:
            return self._data.index[x]
        return None




"""
SFCTOOLS
========

Thomas Baldauf
thomas.baldauf@dlr.de
November 2021, German Aerospace Center

Licensed under MIT license

Contents

    1. CORE FEATURES
    2. DATA STRUCTURES
    3. AUTOMATION AND PARAMETER-IO
    4. BOTTOM-UP FEATURES
    5. MISC
    6. GRAPHICAL USER INTERFACE

"""

"""
CORE FEATURES
"""

from .core.agent import Agent, SingletonAgent
from .core.agent import block_on_bankrupt
from .core.singleton import Singleton


from .core.clock import Clock
from .core.settings import Settings
from .core.world import World
from .core.flow_matrix import FlowMatrix, Accounts


"""
DATA STRUCTURES
"""

from .datastructs.balance import BalanceSheet, BalanceEntry
from .datastructs.income_statement import IncomeStatement,ICSEntry
from .datastructs.cash_flow_statement import CashFlowStatement, CashFlowEntry

from .datastructs.bank_order_book import BankOrderBook
from .datastructs.inventory import Inventory
from .datastructs.market_registry import MarketRegistry
from .datastructs.worker_registry import WorkerRegistry
from .datastructs.signalslot import Signal,Slot,SignalSlot

"""
AUTOMATION AND PARAMETER-IO
"""
from .automation.runner import ModelRunner
# from .automation.calibration import CalibrationRoutine, DistanceMeasure # <- experimental thus commented out

"""
BOTTOM-UP FEATURES
"""

from .bottomup.stock_manager import StockManager
from .bottomup.matching import MarketMatching
# from .bottomup.productiontree import ProductionTree # <- experimental thus commented out

"""
MISCELLANEOUS
"""

# from .misc.timeseries import stretch_pandas, stretch_datetime, convert_quarterly_to_datetime, stretch_to_length, convert_numeric
# from .misc.mpl_plotting import matplotlib_barplot,matplotlib_lineplot, plot_sankey
from .misc.reporting_sheet import ReportingSheet, DistributionReport, IndicatorReport


"""
GUI
"""
# from . import gui



__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '23.06.202'
__status__ = 'dev' # options are: dev, test, prod

from sfctools.examples.example_wrapper import Example

from sfctools.datastructs.worker_registry import WorkerRegistry
from sfctools import Agent
import numpy as np
np.random.seed(1980)

def run():

    class Worker(Agent):
        def __init__(self):
            super().__init__()
            self.reservation_wage = np.random.rand()

    a = Agent()

    wreg = WorkerRegistry(owner=a,wage_attr="reservation_wage") # create a new worker registry
    workers = [Worker() for i in range(10)] # create a bunch of workers

    for worker in workers: # add workers to registry
        wreg.add_worker(worker)
    
    costs = wreg.get_avg_costs() # get average reservation wage of all workers
    print("Average cost: %.2f\n" % costs)

    wreg.fire_random(2) # fire two random workers

    # pop workers from the worker stack
    print("Wokers left:")
    for i in range(8):
        print(wreg.pop())
    

class WorkerExample(Example):
    def __init__(self):
        super().__init__(run)


if __name__ == "__main__":
    my_instance = WorkerExample()
    my_instance.run()


__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '06.10.2022'
__status__ = 'dev' # options are: dev, test, prod

from sfctools.examples.example_wrapper import Example

from sfctools.datastructs.signalslot import Signal,Slot
from sfctools.datastructs.signalslot import SignalSlot
from sfctools import Agent

def run():

    class Sender(Agent): # a dummy receiver
        def __init__(self):
            super().__init__()
            self.message = None


    class Receiver(Agent): # a dummy sender
        def __init__(self):
            super().__init__()
            self.recording = None

    # create two dummy agents
    my_sender = Sender()
    my_receiver = Receiver()
    my_other_receiver = Receiver()

    # create signals and slots
    sender_signal = Signal("Message")
    receiver_slot = Slot("Message")
    sender_signal.connect_to([receiver_slot])

    # link signals and slots to agents
    my_sender.message = sender_signal
    my_receiver.recording = receiver_slot
    my_other_receiver.recording = receiver_slot

    # emit signal
    sender_signal.update("Urgent Message")
    sender_signal.emit(verbose=True)

    # read recording
    print("received:", my_receiver.recording.value())
    print("received (other):", my_other_receiver.recording.value())


class SignalSlotExample(Example):
    def __init__(self):
        super().__init__(run)


if __name__ == "__main__":
    my_example = SignalSlotExample()
    my_example.run()


__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod

from .singleton import Singleton
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
import warnings


class Clock(Singleton):
    """
    The clock of the simulation.
    """

    def __init__(self, t0=datetime.min, dt=relativedelta(months=1)):
        """
        Init a world clock.

        :param t0: optional, starting date of the clock
        :param dt: optional, datetime.timedelta describing the real time step. Default is one month.

        Example

        .. code-block:: python

            from sfctools import Clock

            for i in range(500):
                # simulation steps

                Clock().tick()
                print(Clock().get_time(), Clock().get_real_time())

        """

        if self.do_init: # only init the clock at the first constructor call! See Singleton super class implementation

            self._t0 = t0
            self._t = 0  # the simulation tme
            self._real_time = t0  # the real time as datetime object starting at big bang
            self._dt = dt
            self.set_initialized()

    @property
    def t0(self):
        return self._t0

    def reset(self,verbose=False):
        """
        Resets the clock

        :param verbose: bool (default False), triggers a reset warning if True
        """

        if verbose:
            warnings.warn("Clock has been reset")

        self._t = 0  # the simulation tme
        self._real_time = self._t0  # the real time as datetime object starting at big bang

    def tick(self):
        """
        Increases the clock tick by one
        """
        self._t += 1
        self._real_time += self._dt

    def get_time(self):
        """
        Gets the current simulation time as int
        """
        return self._t

    def get_real_time(self):
        """
        Get the real time (the actual time on the 'non-simulation' timeline) as datetime
        """
        return self._real_time  # < TODO test this



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod


from .singleton import Singleton
from .world import World

from collections import defaultdict
import pandas as pd
import warnings
from enum import Enum
import numpy as np
# import seaborn as sns
import networkx as nx


class Accounts(Enum):
    """
    Enum for account type in flow matrix. Current and capital account is allowed
    """
    KA = 0 # capital account
    CA = 1 # current account


class FlowMatrix(Singleton):
    """
    The flow matrix takes care of the flows between agents and the stock-flow consistency. It has the following structure:


+--------------------------------------------------------------------------------------+
| Flow Matrix                                                                          |
+===============+=========================+==================================+=========+
|               |     Agent 1             |           Agent 2                |  TOTAL  |
+---------------+-----------+-------------+----------------+-----------------+---------+
|               |     CA    | KA          |           CA   | KA              |         |
+---------------+-----------+-------------+----------------+-----------------+---------+
|               |           |             |                |                 |         |
| Flows         |    -x     |             |                |  +x             |   0     |
+---------------+-----------+-------------+----------------+-----------------+---------+
|               |           |             |                |                 |         |
| d(Stocks)     |     +x    |             |                |  -x             |   0     |
+---------------+-----------+-------------+----------------+-----------------+---------+
| TOTAL         |     0     |  0          |            0   |    0            |   0     |
+---------------+-----------+-------------+----------------+-----------------+---------+

    | CA: Current Account, KA: Capital Account.

    As a consistency check, the sum of rows and columns should be zero after the end of each simulation period.
    For more information, see

- Investopedia Article https://www.investopedia.com/ask/answers/031615/whats-difference-between-current-account-and-capital-account.asp
- A nice introduction by Gasselli (OECD) https://www.oecd.org/naec/new-economic-policymaking/grasselli_OECD_masterclass_2019.pdf

    """

    reset_count = 0

    def __init__(self):
        # constructor of the flow matrix

        if hasattr(self, "initialized"):  # only initialize once
            return

        self.initialized = True # initialized flag

        # TODO maybe think about replacing above code with 'if self.do_init:' check?
        # Keeping it for now because it works...

        self._flow_data = {}  # has two keys CA and KA, see reset()...
        self.reset()  # defaultdict will take care of missing keys here
        self.linear_log = [] # linear transaction log

    def reset(self,verbose=False):
        """
        reset the data
        :param verbose: boolean (default False), triggers a reset warning if True
        """

        self._flow_data[Accounts.CA] = defaultdict(lambda: defaultdict(float))
        self._flow_data[Accounts.KA] = defaultdict(lambda: defaultdict(float))

        if self.__class__.reset_count > 1:

            # warn the user every time this is being reset!
            if verbose:
                warnings.warn("FlowMatrix has been reset")

        else:
            self.__class__.reset_count += 1

    def log_flow(self,kind,quantity,agent_from,agent_to,subject,price=None,invert=False):
        """
        Registers a flow at the flow matrix. This is no method for the user in most cases. It is automatically generated in the qattune gui.

        :param kind: tuple (from_account,to_account)
        :param quantity: weight (quantity) of the flow
        :param price: float or None (default). If not None, a price conversion factor is applied. Should not be used in most cases
        :param invert: reverse sign of the transferred quantity? Default False, should not be used in most cases
        :param agent_from: Agent instance (optionally str), sender agent
        :param agent_to: Agent instnace (optionally str), receiver agent

        Example

        .. code-block:: python

            from sfctools import FlowMatrix, Accounts
            CA = Accounts.CA
            KA = Accounts.KA

            flow = (CA,KA) # from account -> to account
            FlowMatrix().log_flow(flow, 42.0, agent1, agent2, subject="my_subject")
            # ...

            df = FlowMatrix().to_dataframe(group=True)
            print(FlowMatrix().to_string(group=True))
            FlowMatrix().check_consistency() # ok if no error is raised

        """

        if price is not None:
            Q = quantity * price
        else:
            Q = quantity

        if invert:
            Q = - Q

        from_account = kind[0]
        to_account = kind[1]

        self._flow_data[from_account][subject][agent_from]-= Q
        self._flow_data[to_account][subject][agent_to] += Q

    def check_consistency(self):
        """
        Checks the consistency of the transaction flow matrix.
        """

        df = self.to_dataframe(group=True)

        if df.empty:
            return

        if not df.empty:
            null_sym = "   .-   "

            # avoid numerical errors
            # so round to at max 4 orders of magntude less than max order of magnitude

            df2 = df.replace(null_sym,0.0).astype(float)
            om_max = int(np.ceil(np.log10(df2.to_numpy().max())))
            om_min = int(np.ceil(np.log10(abs(df2.to_numpy().min()))))
            order_magnitude = max(om_max,om_min)
            df2 = df2.round(-order_magnitude + 4)

            if np.array(df2["Total"]).any():
                raise RuntimeError("Inconsistent Row In Flow Sheet: \n%s" % df.to_string())

            df3 = df2.T
            if np.array(df3["Total"]).any():
                raise RuntimeError("Inconsistent Column In Flow Sheet:\n %s" % df.to_string())

    def to_string(self,group=True):
        """
        converts the flow matrix to string representation
        :return : str
        """
        return self.to_dataframe(group=group).to_string()

    def to_dataframe(self,group =True,insert_nullsym=True):
        """
        Converts the data structure to a human-readable dataframe format.
        WARNING this is slow

        :param group: boolean switch (default True), if True it will group the agents of the same class together
        :param insert_nullsym: will insert '.-' symbol instead of zero, default True
        :return: pandas dataframe object
        """

        df_credit  = pd.DataFrame(self._flow_data[Accounts.CA]).T
        df_capital = pd.DataFrame(self._flow_data[Accounts.KA]).T

        df_merge = pd.concat([df_credit, df_capital], axis=1, keys=['CA', 'KA']).swaplevel(0,1,axis=1).sort_index(axis=1)

        df = df_merge.fillna(0.0).sort_index()

        agent_types = World().get_agent_types()

        if not group:

            df.loc["Total"] = df.sum()
            df["Total"] = df.T.sum()
            return df.round(4)

        else:
            """
            Provide an overview with aggregated classes
            """

            data = {}
            renamer = {}
            for a in agent_types:
                my_group = []
                for b in df.columns:
                    # print("b",b,"a",a,b[0])
                    if isinstance(b[0],a):
                        my_group.append(b[0])

                if len(my_group) > 0:
                    my_df = df.loc[:,df.columns.get_level_values(0).isin(my_group)]
                    my_CA = my_df.loc[:,my_df.columns.get_level_values(1).isin({"CA"})].sum(axis=1)
                    my_KA = my_df.loc[:,my_df.columns.get_level_values(1).isin({"KA"})].sum(axis=1)

                    concat =  pd.concat([my_CA, my_KA], axis=1, keys=['CA', 'KA'])

                    data[a] = concat

                if len(my_group) > 1:
                    plural_suffix = ""

                    if len(World().get_agents_of_type(a.__name__))> 1:
                        plural_suffix = "s"

                    renamer[a] = a.__name__ + plural_suffix
                else:
                    renamer[a] = a.__name__

            if data == {}:
                return pd.DataFrame()

            df2 = pd.concat(data, axis=1)

            df2 = df2.rename(columns=renamer).sort_index()

            # df2 = df2.round(4)

            df2.loc["Total"] = df2.sum()
            df2["Total"] = df2.T.sum()

            df2 = df2.round(5)
            df2 = df2.round(4) # cut one digit to obtain a 'consistently rounded table'

            if insert_nullsym:
                null_sym = "   .-   "
                df2 = df2.replace(0.0,null_sym)

            sorted_cols2 = sorted(df2.columns)
            try:
                df2 = df2.reindex(sorted_cols2, axis=1)

            except ValueError: # non-unique multi-indices here? bad sign... try to remove the non-unique part
                try:
                    sorted_cols2 = sorted(df2.columns)
                    sorted_cols2 = list(set(sorted_cols2)) # try to remove non-unique multi-indices
                    df2 = df2.reindex(sorted_cols2, axis=1)

                    warnings.warn("Encountered non-unique multi-index in FlowMatrix. Something might be wrong")

                except: # other error?
                    raise RuntimeError("Something went wrong when aggregating the FlowMatrix as dataframe. You might be able to fix this error by cross-checking the naming of your agents and transactions.")

            cols = list(df2.columns.values) #Make a list of all of the columns in the df
            #print("COLS",cols)
            cols.pop(cols.index(("Total",""))) #Remove Total from list
            df2 = df2[cols+[("Total","")]] #Create new dataframe with columns in the order you want

            return df2

    def plot(self,group=True):
        self.plot_colored(show_plot=True,group=group)

    def plot_colored(self,show_plot=True,group=True,cmap='coolwarm',show_values=True):
        """
        Plots the flow matrix as a nice colored heat map.
        This will open up a matplotlib window...

        :param show_plot: show the plot as window? default True. If False, figure is returned instead
        :param group: aggregated view?
        :param cmap: colormap to use (default 'coolwarm')
        :param show_values: (default True) show numerical values above colored tiles
        :return fig: figure object
        """
        import matplotlib.pyplot as plt
        # TODO nicer formatting
        # TODO more plotting options
        import seaborn as sns
        df = self.to_dataframe(insert_nullsym=False,group=group)

        fig = plt.figure(figsize=(5,3))
        sns.heatmap(df, annot=show_values ,cmap=cmap,)
        ax = plt.gca()
        ax.set_ylabel('')
        ax.set_xlabel('')

        plt.tight_layout()
        if show_plot:
            plt.show()

        return fig

    def convert_sankey_df(self):
        """
        Converts the FlowMatrix data to a sankey-plottable dataframe format.
        This is meant to increase the flexibility in data retrieval and plotting capabilities. To directly plot the FlowMatrix, see plot_sankey

        :return: tuple of dataframes - sankey_source, sankey_sink
        """

        df = FlowMatrix().to_dataframe(insert_nullsym=False)

        df_CA = df.iloc[:, df.columns.get_level_values(1)=='CA']
        df_CA.columns = df_CA.columns.droplevel(1) # current account

        df_KA = df.iloc[:, df.columns.get_level_values(1)=='KA'] # capital account
        df_KA.columns = df_KA.columns.droplevel(1)

        source  = []
        target = []
        types = []
        value = []

        source2  = []
        target2 = []
        types2 = []
        value2 = []

        subjects = {}
        for i,k in enumerate(df.index):
            subjects[str(k)] = i

        for i, df_i in enumerate([df_CA,df_KA]):

            for index, row in df_i.iterrows():

                for column in df_i.columns:

                    val = float(row[column])
                    if val < 0.0 and index != "Total" and (not index.startswith("Δ")): # is a source
                        source.append(str(column) + "$[source]")
                        target.append(str(index) )
                        value.append(-val)
                        types.append(subjects[str(index)])

                    elif val > 0.0 and index != "Total" and (not index.startswith("Δ")): # is a sink
                        source2.append(str(column) + "$[sink]")
                        target2.append(str(index) )
                        value2.append(val)
                        types2.append(subjects[str(index)])

        my_sankey_source = pd.DataFrame({"from":source,
                                  "to":target,
                                  "color_id":types,
                                  "value":value}).round(2)
        my_sankey_sink =  pd.DataFrame({"to":source2,
                                  "from":target2,
                                  "color_id":types2,
                                  "value":value2}).round(2)

        return my_sankey_source, my_sankey_sink

    def plot_sankey(self,show_values=True, show_plot=True,colors=None,label_rot=0.0,filling_fraction=0.85):
        """
        plots a sankey diagram of the flow matrix

        :param show_values: boolean switch to plot the values of the edge weights
        :param show_plot: show the plot as window? default True. If False, figure is returned instead.
        :param colors: (optional) a list of colors (if None, default colors are chosen)
        :param label_rot: rotation of the labels 
        :param filling_fraction: controls the 'thickness' of the Sankey bands

        :return fig: figure object
        """
        import matplotlib.pyplot as plt
        from ..misc.mpl_plotting import plot_sankey as sankey

        my_sankey_source,my_sankey_sink = self.convert_sankey_df()
        fig = sankey([my_sankey_source,my_sankey_sink],show_values=show_values,show_plot=show_plot,colors=colors,label_rot=label_rot,filling_fraction=filling_fraction)
        return fig


    @property
    def capital_flow_data(self):
        """'shortcut' property. only returns the data from the capital account"""
        return self._flow_data[Accounts.KA]

    @property
    def current_flow_data(self):
        """'shortcut' property. only returns the dta fromt he current flow account"""
        return self._flow_data[Accounts.CA]




from scipy.interpolate import interp1d
import numpy as np 
import matplotlib.pyplot as plt 


def my_interpolate(mypoints,method,plot=False,n = 45):

    #p0 = points[0]
    #points = points + [(p0[0],p0[1])]

    points = [(p[0],p[1]) for p in mypoints]
    #print("points", points)

    distance = np.cumsum( np.sqrt(np.sum( np.diff(points, axis=0)**2, axis=1 )) )
    distance = np.insert(distance, 0, 0)

    if len(distance) > 0:
        distance = distance/distance[-1]
    
    # method = "square"
    alpha = np.linspace(0,1,n)
    interpol =  interp1d(distance, points, kind = method , axis=0)
    interpolated_points = interpol(alpha)

    if plot:
        plt.figure()
        plt.scatter([p[0] for p in points], [p[1] for p in points], facecolor="white", color="black")
        plt.scatter([p[0] for p in interpolated_points], [p[1] for p in interpolated_points], color="red")
        plt.show()

    return interpolated_points

if __name__ == "__main__":
    points = [(0,4), (5,5), (1,8)]

    my_interpolate(points,method="quadratic",plot=True)

    points = [(0,4), (0,5), (0,6), (0,7)]
    my_interpolate(points,method="quadratic",plot=False)


__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '06.10.2022'
__status__ = 'dev' # options are: dev, test, prod


from sfctools import Agent, BalanceSheet, BalanceEntry
import numpy as np
import matplotlib.pyplot as plt

from sfctools.examples.example_wrapper import Example


def run(show_figure=True):

    class MyAgent(Agent):
        """
        An example agent
        """
        def __init__(self):
            super().__init__()

            self.change_equity(20.0,"Asset1")
            self.change_equity(10.0,"Asset2")
            self.change_debt(18.0)

        def change_equity(self,q,which): # increase or decrease equity by q / -q
            with self.balance_sheet.modify:
                self.balance_sheet.change_item(which,BalanceEntry.ASSETS,q)
                self.balance_sheet.change_item("Equity",BalanceEntry.EQUITY,q)

        def change_debt(self,q): # increase or decrease debt by q / -q
            with self.balance_sheet.modify:
                self.balance_sheet.change_item("Debt",BalanceEntry.LIABILITIES,q)
                self.balance_sheet.change_item("Asset2",BalanceEntry.ASSETS,q)

        def swap_assets(self,q): # swap between assets
            try:
                with self.balance_sheet.modify:
                    self.balance_sheet.change_item("Asset2",BalanceEntry.ASSETS,-q)
                    self.balance_sheet.change_item("Asset1",BalanceEntry.ASSETS,q)
            except:
                pass

        def update(self):

            q1 = 0.5*(np.random.rand())
            q2 = 0.5*(np.random.rand())
            q3 = .4*np.random.rand()

            self.change_equity(q1,"Asset1")
            self.change_debt(q1)
            self.swap_assets(q3)

    my_agent = MyAgent() # create an agent

    my_balances = [] # allocate empty list

    for i in range(80): # iterate simulation
        my_agent.update()
        my_balances.append(my_agent.balance_sheet.raw_data) # append balance sheet data to list

    # plot balance sheet evolution
    if show_figure:
        BalanceSheet.plot_list(my_balances, dt=5, xlabel="Time", title="My Agent", show_liabilities=True)





class BalanceSheetExample(Example):
    def __init__(self):
        super().__init__(lambda: run(show_figure=False))


if __name__ == "__main__":
    my_instance = BalanceSheetExample()
    my_instance.run(show_figure = True)



__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '06.10.2022'
__status__ = 'dev' # options are: dev, test, prod

from sfctools.examples.example_wrapper import Example

from sfctools import Settings, Agent, FlowMatrix, BalanceEntry, World
from sfctools import Accounts

import os

def run():

    print(os.getcwd())

    try:
      Settings().read("sfctools/examples/basic_example/my_settings.yml") # <- defines parameter 'beta'
    except:
      Settings().read("my_settings.yml") # <- defines parameter 'beta'

    print(Settings())

    FlowMatrix().reset()
    World().reset()

    class MyAgent(Agent):
      def __init__(self):
        super().__init__()

        self.my_parameter = Settings().get_hyperparameter("beta") # or Settings()["beta"]

      # def more_here(self,*args):
      #  ...

    my_agent = MyAgent() # <- create an agent
    my_second_agent = MyAgent()  # <- create a second agent

    with my_agent.balance_sheet.modify:
      my_agent.balance_sheet.change_item("Cash", BalanceEntry.ASSETS, 10.0)
      my_agent.balance_sheet.change_item("Equity", BalanceEntry.EQUITY, 10.0)  # enlarge my_agent's balance by 10

    def my_test_transaction(agent1,agent2,quantity):
      FlowMatrix().log_flow((Accounts.CA, Accounts.CA), quantity, agent1,agent2,subject="test")

    my_test_transaction(my_agent,my_second_agent,9.0) # transfer 9 units between the agents

    print(my_agent.balance_sheet.to_string())
    print(my_second_agent.balance_sheet.to_string())

    print(FlowMatrix().to_string(group=False))


class BasicExample(Example):
    def __init__(self):
        super().__init__(run)

if __name__ == "__main__":
    my_instance = BasicExample()
    my_instance.run()



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod


from sfctools import SingletonAgent, Settings, World,Clock
from .transactions import wag as transaction_wages 
from .transactions import tax as transaction_taxes 
import pandas as pd 
import numpy as np


class Production(SingletonAgent):
    """
    The production agent produces a single good Y
    """

    def __init__(self):
        """
        constructor for the production agent
        """
        super().__init__(alias="Production") # there is only one production

        self.Y = 1.0 # initial production level 
        self.W = 1.0 # initial wages
        self.T = 1.0 # initial taxes

        gdp_file = Settings().get_hyperparameter("gdp_change")
        self.Y_change = np.loadtxt(gdp_file) # percentage change of GDP , in percent

        self.theta = Settings().get_hyperparameter("theta")

    def link(self):
        """
        Links the production agent with the household agent and the government agent
        """

        self.household = World().get_agents_of_type("Household")[0]
        self.government = World().get_agents_of_type("Government")[0]

    def reset(self):
        """
        Resets the agent for the next simulation period
        """
        self.income_statement.reset()

    def update(self):
        """
        updates production 
        """       
        # set the wage level for this period
        
        t = Clock().get_time()
        C = self.household.C
        G = self.government.G

        self.W = C + G # update wage level 

        # pay wages 
        wages = (1-self.theta) * self.W
        transaction_wages(self,self.household,wages)

        # pay taxes
        taxes = self.theta * self.W
        self.T = taxes 
        transaction_taxes(self,self.government,taxes) # production pays income taxes of households directly to household

        self.Y = self.W - self.T
        self.Y *= (1+0.01*self.Y_change[t]) 


    def file_bankruptcy(self,event):
        pass


__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'test' # options are: dev, test, prod

from collections import defaultdict
from ..core.agent import Agent
from ..bottomup.matching import MarketMatching
import pandas as pd
import numpy as np


class BankOrderBook:
    """
    The order book keeps track of all agents that have deposits and loans at a bank.

    loans data:

    .. code-block:: python

            {
            Agent1: list[Table],
            Agent2: list[Table],
            ...
            }

    Loans:

    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
    |  ID      | Agent     | Outstanding Loans | interest on loans (r_l) |  Principal                      | Time | Maturity       |
    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
    |     0    |  10.0     | 10.0              | 0.05                    |  100.0                          | ...  |                |
    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
    |     1    |   ...     |                   |                         |                                 |      |                |
    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
    |          |           |                   |                         |                                 |      |                |
    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
    |          |   ...     |                   |                         |                                 |      |                |
    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
    |          |           |                   |                         |                                 |      |                |
    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+
    |     42   |   ...     |                   |                         |                                 |      |                |
    +----------+-----------+-------------------+-------------------------+---------------------------------+------+----------------+


    Deposits:

    +----------+-----------+-----------+-----------------------------+---------------------------------+------------+
    |  ID      | Agent     |  Depos.   | interest on deposits (r_d)  |   Time                          |  Duration  |
    +----------+-----------+-----------+-----------------------------+---------------------------------+------------+
    |     0    |  10.0     |  0.05     |                             |  0.01                           |            |
    +----------+-----------+-----------+-----------------------------+---------------------------------+------------+
    |     1    |   ...     |           |                             |                                 |            |
    +----------+-----------+-----------+-----------------------------+---------------------------------+------------+
    |          |           |           |                             |                                 |            |
    +----------+-----------+-----------+-----------------------------+---------------------------------+------------+
    |          |   ...     |           |                             |                                 |            |
    +----------+-----------+-----------+-----------------------------+---------------------------------+------------+
    |          |           |           |                             |                                 |            |
    +----------+-----------+-----------+-----------------------------+---------------------------------+------------+
    |     42   |   ...     |           |                             |                                 |            |
    +----------+-----------+-----------+-----------------------------+---------------------------------+------------+

    """

    def __init__(self, owner,credit_market:MarketMatching=None,deposit_market:MarketMatching=None):
        """
        constructor

        :param owner: agent instance who owns this data structure
        :credit_market: a MarketMatching obect for the credit market (default None). Providing an instance will automatically unlink agents when the deposit is dissolved.
        :deposit_market:  a MarketMatching object or the deposit market (default None). Providing an instance will automatically unlink agents when the deposit is dissolved.

        """

        self.owner = owner
        self.loans_data = defaultdict(list)

        """
        defaultdata of loans
        """

        self.deposits_data = defaultdict(list)

        """
        defaultdict of deposits
        """

        self.idx_d = 0  # index counter deposits
        self.idx_l = 0  # index counter loans

        self.total_deposits = 0.0
        self.total_loans = 0.0

        self.credit_market = credit_market
        self.deposit_market = deposit_market

    def __repr__(self):
        return "<Order Book of %s>" %self.owner

    def to_dataframe(self):
        """
        returns pandas datafame representation of this data struture

        :return: deposits_df, loans_df (tuple)
        """

        data = {"ID":[],"Agent":[],"Outstanding Loans":[],"Principal": [], "r_l":[],"Time":[],"Maturity":[]}
        for agent, listing in self.loans_data.items():
            for entry in listing:
                data["ID"].append(entry["idx"])
                data["Agent"].append(agent)
                data["Outstanding Loans"].append(entry["Outstanding Loans"])
                data["r_l"].append(entry["r_l"])
                data["Principal"].append(entry["Principal"])
                data["Time"].append(entry["Time"])
                data["Maturity"].append(entry["Maturity"])

        loans_df = pd.DataFrame(data).set_index("ID")

        data = {"ID": [], "Agent":[], "Deposits": [], "r_d": [], "Time":[],"Duration":[]}

        for agent,listing in self.deposits_data.items():
            for entry in listing:
                data["ID"].append(entry["idx"])
                data["Agent"].append(agent)
                data["Deposits"].append(entry["Deposits"])
                data["r_d"].append(entry["r_d"])
                data["Time"].append(entry["Time"])
                data["Duration"].append(entry["Duration"])

        deposits_df =  pd.DataFrame(data).set_index("ID")

        return deposits_df,loans_df

    def to_string(self):
        """
        string representation of this data structure
        """
        s = "----  ORDER BOOK of {0:20} -----------------------------------------".format(str(self.owner))

        deposits_df,loans_df = self.to_dataframe()

        if loans_df.empty:
            loans_str = "(None)"
        else:
            loans_str = loans_df.to_string()

        s+= "\nLoans: %.2f\n"%self.total_loans + loans_str+ "\n\n"

        if deposits_df.empty:
            deposits_str = "(None)"
        else:
            deposits_str = deposits_df.to_string()

        s += "\nDeposits: %.2f\n"%self.total_deposits +deposits_str+ "\n\n"
        s += "----------------------------------------------------------------------------------"
        return s

    def remove_loan(self, idx, agent):
        """
        Removes a whole entry (this should be called when a loan is entirely paid back)

        :param idx: id of the loan
        :param agent: agent belonging to the loan
        """

        if self.credit_market is not None:
            self.credit_market.unlink_agents(supply_agent=self.owner, demand_agent=agent)

        remove_index = None
        # find the entry with the corresponding index
        for i,entry in enumerate(self.loans_data[agent]):
            if entry["idx"] == idx:
                remove_index = i

        if remove_index is None:
            raise KeyError("Could not find entry")

        self.total_loans -= self.loans_data[agent][remove_index]["Outstanding Loans"]
        del self.loans_data[agent][remove_index]

    def remove_deposit(self, idx, agent):
        """
        remove a deposit by index

        :param idx: index of entry
        :param agent: agent who belongs to this deposit
        """

        if self.deposit_market is not None:
            self.deposit_market.unlink_agents(supply_agent=agent, demand_agent=self.owner)

        remove_index = None
        # find the entry with the corresponding index
        for i, entry in enumerate(self.deposits_data[agent]):
            if entry["idx"] == idx:
                remove_index = i

        if remove_index is None:
            raise KeyError("Could not find entry")

        self.total_deposits -= self.deposits_data[agent][remove_index]["Deposits"]

        # print("REMOVE DEPOSIT", self, idx, agent,self.deposits_data[agent][remove_index]["Deposits"])

        del self.deposits_data[agent][remove_index]

    def withdraw_from_account(self,agent,quantity):
        """
        BETA withdraw money from deposit account.

        :param agent: agent who withdraws
        :param quantity: quantity to withdraw
        """

        # print("WITHDRAW DEPOSIT", self,  agent,quantity)

        x = 0
        removals = []

        for i, entry in enumerate(self.deposits_data[agent]):
            D = entry["Deposits"]
            idx = entry["idx"]

            remaining_amount = quantity-x
            if remaining_amount < D:  # only change the account
                self.deposits_data[agent][i]["Deposits"] -= remaining_amount

                assert self.deposits_data[agent][i]["Deposits"] >= 0, print(self.deposits_data[agent][i]["Deposits"])
                self.total_deposits -= remaining_amount
                x += remaining_amount

                break

            else:  # completely remove the account
                removals.append(idx)
                x += D

        for idx in removals:
            self.remove_deposit(idx, agent)


    def add_deposits(self, agent: Agent, new_deposits, r_d,t=np.inf):
        """
        add deposits of an agent at this bank

        :param agent: reference to agent
        :param new_deposits: amount of new deposits
        :param r_d: interest rate on deposits
        :param t: time of the deposits
        """

        # print("ADD DEPOSIT", self, agent, new_deposits)
        # self.depsosit_market.link_agents(supply_agent=agent, demand_agent=self.owner,r_d) < this should be done inside market matching

        self.idx_d += 1

        self.deposits_data[agent].append({
            "idx":self.idx_d,
            "Deposits": new_deposits,
            "r_d":r_d,
            "Time":0,
            "Duration":t
        })

        self.total_deposits += new_deposits


    @property
    def depositors(self) -> dict:
        # gets a list of depositors and their deposits

        deps = defaultdict(lambda: 0.0)
        for key,val in self.deposits_data.items():

            agent = key
            x = 0.0
            for entry in val:
                x += entry["Deposits"]

            deps[agent] = x

        return deps

    @property
    def debtors(self) -> dict:
        # gets a list of debtors and their loans

        loans = defaultdict(lambda: 0.0)
        for key, val in self.loans_data.items():

            agent = key
            x = 0.0
            for entry in val:
                x += entry["Outstanding Loans"]

            loans[agent] = x

        return loans


    def get_deposit_interest_of(self,agent) -> float:
        """
        compute the mean interest rate an agent is paid
        """

        i = []
        for entry in self.deposits_data[agent]:
            i.append(entry["r_d"])
        return np.mean(i)

    def get_loan_interest_of(self, agent) -> float:
        """
        compute the mean interest rate an agent pays

        :param agent: reference to an agent instance
        """

        i = []
        for entry in self.deposits_data[agent]:
            i.append(entry["r_l"])
        return np.mean(i)

    def get_tot_deposits_of(self, agent) -> float:
        """
        returns the total deposts of an agent

        :param agent: reference to an agent instance
        """

        total_deposits = 0.0
        for entry in self.deposits_data[agent]:
            total_deposits += entry["Deposits"]

        return total_deposits

    def get_tot_loans_of(self,agent) -> float:
        """
        returns the total loans of an agent

        :param agent: reference to an agent instance
        """

        total_loans = 0.0
        for entry in self.loans_data[agent]:
            total_loans += entry["Loans"]

        return total_loans

    def get_loans_of(self,agent):
        """
        get a list of dicts with the information on loans for an agent

        :param agent: the reference agent
        :return: list of dicts
        """

        return self.loans_data[agent]

    def get_deposits_of(self,agent):
        """
        get a list of dicts with the information on deposits for an agent

        :param agent: the reference agent
        :return: list of dicts
        """

        return self.deposits_data[agent]

    def add_loans(self, agent:Agent, new_loans:float, r_l: float, t=20):
        """
        add loans of an agent at this bank

        :param agent: reference to agent
        :param new_loans: amount of new loans
        :param r_l: interest rate on loans
        :param t: payback time (default 20 periods)
        """

        assert new_loans > 0

        # self.credit_market.link_agents(supply_agent=self.owner, demand_agent=agent,r_l) <- this should be done inside market matching

        self.idx_l += 1

        self.loans_data[agent].append({
            "idx": self.idx_l,
            "Outstanding Loans": new_loans,
            "r_l": r_l,
            # "payback": new_loans/(1.0*t), DEPRECATED TB 23/11/22
            "Principal": new_loans,
            "Time": 0,
            "Maturity": t
        }) # data = {"ID":[],"Agent":[],"Outstanding Loans":[],"Principal": [], "r_l":[],"Time":[],"Maturity":[]}

        self.total_loans += new_loans


    def debt_equity_swap(self, agent, q):
        """
        BETA writes off a debt-equity swap

        :param agent: refrence to an agent instance
        :param q: quanity to write off
        """

        entries = self.loans_data[agent]

        cumu = 0  # cumulative variable (debt written off)

        for i, entry in enumerate(entries):

            if cumu >= 0:
                break

            if entry["Outstanding Loans"] <= q - cumu:  # require whole entry?
                cumu += entry["Outstanding Loans"]
                self.remove_loan(entry["idx"], agent)

            else:  # require only part of entry?
                cumu += q - cumu
                self.total_loans -= q - cumu

                self.loans_data[agent][i]["Outstanding Loans"] -= q - cumu  # partially reduce entry

                # re-adjust the amount of loan payback
                # t_remain = self.loans_data[agent][i]["Principal"] - self.loans_data[agent][i]["time"]
                # self.loans_data[agent][i]["payback"] = self.loans_data[agent][i]["Loans"] / (1.0 * t_remain)
                # DEPRECATED TB 23/11/22

                return

    def write_off_bad_debt(self, agent, q):
        """
        BETA books a bad debt. This means that loans are written off this data struct.

        :param agent: the agent concerning the bad debt
        :param q: amount of bad debt to write off
        """

        assert q >= 0,  q

        entries = self.loans_data[agent]

        cumu = 0  # cumulative variable (debt written off)

        for i, entry in enumerate(entries):

            if cumu >= q:
                break

            if entry["Outstanding Loans"] <= q-cumu:  # require whole entry?
                self.remove_loan(entry["idx"],agent)
                cumu += entry["Outstanding Loans"]

            else:  # require only part of entry?
                self.loans_data[agent][i]["Outstanding Loans"] -= q-cumu  # partially reduce entry
                cumu += q-cumu
                self.total_loans -= q-cumu

                # re-adjust the amount of loan payback
                # t_remain = self.loans_data[agent][i]["Maturity"] - self.loans_data[agent][i]["time"]
                # self.loans_data[agent][i]["payback"] = self.loans_data[agent][i]["Loans"] / (1.0*t_remain)
                # DEPRECATED TB 23/11/22

                return



__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '06.10.2022'
__status__ = 'dev' # options are: dev, test, prod

from sfctools.examples.example_wrapper import Example


def run():

    from sfctools import Agent , BalanceEntry

    class MyAgent(Agent):

        def __init__(self):

            super().__init__()

            with self.balance_sheet.modify:
                self.balance_sheet.change_item("Cash", BalanceEntry.ASSETS, 10.0)
                self.balance_sheet.change_item("Equity", BalanceEntry.EQUITY, 10.0)

            print("Hello Agent")


    my_agent = MyAgent()

    print(my_agent.balance_sheet.to_string())



class HelloWorldExample(Example):
    def __init__(self):
        super().__init__(run)



__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '06.10.2022'
__status__ = 'dev' # options are: dev, test, prod

from sfctools.examples.example_wrapper import Example
from sfctools import Agent, FlowMatrix, Accounts, BalanceEntry, World

def run(show_plot=True):

    FlowMatrix().reset()
    World().reset()

    # create some agent classes
    class MyAgent(Agent):
     def __init__(self):
         super().__init__()
         with self.balance_sheet.modify:

             self.balance_sheet.change_item("Cash",BalanceEntry.ASSETS,+100.0,suppress_stock=True)
             self.balance_sheet.change_item("Equity",BalanceEntry.EQUITY,+100.0,suppress_stock=True)


    class TypeA(MyAgent):
     def __init__(self):
         super().__init__()

    class TypeB(MyAgent):
     def __init__(self):
         super().__init__()

    class TypeC(MyAgent):
     def __init__(self):
         super().__init__()

    class TypeD(MyAgent):
     def __init__(self):
         super().__init__()

    class TypeE(MyAgent):
     def __init__(self):
         super().__init__()


    my_a = TypeA() # create agents
    my_b = TypeB()
    my_c = TypeC()
    my_d = TypeD()
    my_e = TypeE()
    my_e2 = TypeE()

    def transfer(t,agent1,agent2,subject,quantity): # define a generic transaction
     with agent1.balance_sheet.modify:
         with agent2.balance_sheet.modify:

             agent1.balance_sheet.change_item("Cash",BalanceEntry.ASSETS,-quantity)
             agent1.balance_sheet.change_item("Equity",BalanceEntry.EQUITY,-quantity)

             agent2.balance_sheet.change_item("Cash",BalanceEntry.ASSETS,+quantity)
             agent2.balance_sheet.change_item("Equity",BalanceEntry.EQUITY,+quantity)

             FlowMatrix().log_flow(t, quantity, agent1, agent2 ,subject=subject)

    # do some transactions
    transfer((Accounts.CA, Accounts.CA),my_a,my_b,"my subject",quantity=42)
    transfer((Accounts.CA, Accounts.CA),my_a,my_c,"other subject",quantity=3.1415)
    transfer((Accounts.CA, Accounts.KA),my_c,my_d,"different subject",quantity=10.0)
    transfer((Accounts.KA, Accounts.KA),my_a,my_e,"no subject",quantity=3.1415)
    transfer((Accounts.KA, Accounts.KA),my_a,my_e2,"no subject",quantity=3.1415)

    # output flow matrix as table and Sankey chart
    if show_plot:
        print(FlowMatrix().to_dataframe())
        FlowMatrix().plot_colored(group=False)
        FlowMatrix().plot_colored(group=True)
        FlowMatrix().plot_sankey()


class FlowMatrixExample(Example):
    def __init__(self):
        super().__init__(lambda: run(show_plot=False))


if __name__ == "__main__":
    my_instance = FlowMatrixExample()
    my_instance.run()



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod


import time
from ..datastructs.collection import Collection
from .clock import Clock
from .singleton import Singleton
from ..core.custom_warnings import warning_on_one_line

import warnings
import pandas as pd

# from ..misc.mpl_plotting import matplotlib_barplot,matplotlib_lineplot
import sys
import gc



class World(Singleton):
    """
    This is the world where agents live in. There can only be one world at a time.
    """

    """
    # TODO for the future
    # TODO employ a world.plot() method to get a graph view (or multiple graph views) of the world with networkx(?)
    # TODO world info printer (number of agents, current time (clock status), stocks,...)

    """

    def __init__(self):
        """
        The world constructor.
        """

        if hasattr(self, "initialized"):  # only initialize once
            return
        self.initialized = True

        self.agent_registry = Collection(kind="list")  # register the agents in a data structure
        self.helper_registry = Collection(kind="list") # register the helper structures here

        self.globals = {} # globally accessible data here
        self.agent_types = []

        warnings.formatwarning = warning_on_one_line


    def get_agent_types(self):
        """
        Constructs a list of all agent types registered so far (i.e. keys of agent registry dict).

        :return list: list of agents
        """
        return self.agent_types # return list directly

    def link(self,verbose=False):
        """
        Calls the 'link' fun in all the agents to inter-link them in the world
        """
        pool_a = list(self.helper_registry.values())
        pool_b = list(self.agent_registry.values())

        #if verbose:
        #    print("[WORLD] LINK ... %i agents" % (len(pool_a)+len(pool_b)))

        for agent_list in pool_a + pool_b:
            for agent in list(set(agent_list)):
                if hasattr(agent,"link"):
                    if callable(getattr(agent, "link")):
                        agent.link()
                        if verbose:
                            print("[WORLD] LINK", agent )
                    else:
                        if verbose:
                            warnings.warn("[WORLD] LINK: Link attr of %s is not callable. Skipping..." % str(agent))
                else:
                    if verbose:
                        warnings.warn("[WORLD] LINK: %s has no link attr." % agent)

                    # NOTE it is ok if it has no link attr

    def reset(self,verbose=False):
        """
        resets the registry

        :param verbose: boolean (default False), triggers a reset warning if True
        """

        if verbose:
            warnings.warn("WORLD RESET - deleting all agents.")

        for agent_list in self.agent_registry.values():
            for agent in agent_list:
                try:
                    self.remove_agent(agent)
                    del agent
                except Exception as e:
                    warnings.warn(str(e))


        for elements in self.helper_registry.values():
            for element in elements:
                try:
                    self.remove_helper(element)
                    del element
                except Exception as e:
                    warnings.warn(str(e))

        self.agent_registry = Collection(kind="list")   # register the agents in a data structure
        self.helper_registry = Collection(kind="list")  # register the helpers in a data structure

        clock = Clock()
        if clock is not None:
            clock.reset()

        gc.collect()

    def register_agent(self, agent, verbose=False):
        """
        inserts a new agent under the tag of its class

        :param agent: an agent object to be registered
        """
        if verbose:
            print("[WORLD] Reigster Agent %s of class %s " % (agent,agent.__class__))

        if agent.__class__ not in self.agent_types:
            self.agent_types.append(agent.__class__) # append new key if doesnt exist yet

        self.agent_registry.insert(agent.__class__.__name__, agent) # write new agent to data structure

    def register_helper(self, object):
        """
        inserts a new helper object under the tag of its class

        :param object: instance to store
        """
        self.helper_registry.insert(object.__class__.__name__, object) # write new object

    def find_agent(self,name):
        """
        search for agent by name and alias

        :return: agent instance
        """

        for agent_list in self.agent_registry.values():
            for agent in agent_list:
                if agent.name == name or agent.alias == name:
                    return agent

        return None

    def remove_agent(self, agent, verbose=False):
        """
        removes an agent from the tag of its class (i.e. from the world)

        :param agent: instance to be removed
        """
        if verbose:
            print("[WORLD] Remove Agent %s of class %s " % (agent, agent.__class__))

        if len(self.agent_registry[agent.__class__.__name__]) <= 1 and agent.__class__ in self.agent_types:
            # last agent? -> remove whole class from registry
            self.agent_types.remove(agent.__class__)

        self.agent_registry.remove(agent.__class__.__name__, agent)  # remove the agent from the tag
        del agent  # delete the instance to assure proper removal of the agent

    def remove_helper(self, agent):
        """
        removes an agent from the tag of its class (i.e. from the world)

        :param agent: instance to be removed
        """
        #self.agent_types.remove(agent.__class__)
        self.helper_registry.remove(agent.__class__.__name__, agent)  # remove the agent from the tag
        del agent  # delete the instance to assure proper removal of the agent

    def write_global(self,key,data):
        """
        write globally accessible data here

        :param key: key for the data (generic)
        :param data: value of the data (generic)
        """
        self.globals[key] = data

    def read_global(self,key):
        """
        reads global variable by key
        """
        return self.globals[key]

    def get_agents_of_type(self,agent_type):
        """
        gets all agents belonging to a certain tpe

        :param agent_type: key to search for in self.agent_registry
        :return: list of instances or empty list

        Example

        .. code-block:: python

            from sfctools import Agent,World

            class MyAgent(Agent):
                def __init__(self):
                    super().init()
                    # ...

            [MyAgent() for i in range(10)] # create 10 agents

            mylist = World().get_agents_of_type("MyAgent")
            print(mylist)

        """

        return self.agent_registry[agent_type]



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'prod' # options are: dev, test, prod


class Singleton(object):
    """
    Plain vanilla singleton class
    """
    _instance = None
    __initialized = False

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Singleton, cls).__new__(cls) # , *args, **kwargs)
        return cls._instance

    @property
    def do_init(self):
        return not self.__initialized

    def set_initialized(self):
        self.__initialized = True



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'test' # options are: dev, test, prod

from collections import defaultdict
from ..core.clock import Clock
import pandas as pd
from ..core.settings import Settings
from ..datastructs.balance import BalanceEntry


class Inventory:
    """
    Inventory data structure.
    An Inventory consists of a collection of current assets.

    +-----------------+--------------------------------+-------------------------+
    |   Name of asset | inventory (in physical quant.) | costs ( = worth in €)   |
    +=================+================================+=========================+
    |   Apples        | 10                             | 2.0                     |
    +-----------------+--------------------------------+-------------------------+
    |   Oranges       | 9                              | 2.1                     |
    +-----------------+--------------------------------+-------------------------+

    Usage: The inventory is a non-standard feature of an agent and has to be added manually

    .. code-block:: python
    
        class InventoryAgent(Agent):
            def __init__(self):
                super().__init__()
                self.inventory = Inventory(self)
            
        my_agent = InventoryAgent()
        my_agent.inventory.add_item(name,quantity,costs)
        my_agent.inventory.remove_item(name, quantity)
        my_agent.inventory.get_inventory(name)
        
    """

    def __init__(self, owner,depreciation_dict=None):
        """
        Instantiates a new inventory data structure.

        :param owner: instance who owns this data structure
        :param depreciation_dict: a dict with the depreciation values for the goods stored in the
        """

        self.owner = owner
        self.data = defaultdict(lambda: {"inventory": 0.0, "costs": 0.0})

        if depreciation_dict is not None:
            self.depreciation_dict = depreciation_dict
        else:
            self.depreciation_dict = Settings().depr_dict

    def to_dataframe(self):
        """
        converts the inventory data to pandas dataframe format
        """

        data = {"Asset":[],"Inventory":[], "Worth":[]}

        for asset,line in self.data.items():

            data["Asset"].append(asset)
            data["Inventory"].append(line["inventory"])
            data["Worth"].append(line["costs"])

        return pd.DataFrame(data)

    def to_string(self):
        """
        converts the inventory data to string format
        """

        return "\n\n" + self.to_dataframe().to_string() + "\n\n"

    def depreciate(self, method="linear"):
        """
        Depreciates the values in the balance sheet. This requires a sfctools.Settings object to be instantiated.
        
        :param method: 'linear' (other options not yet implemented)
        """
        if method != "linear":
            raise NotImplementedError("Not yet implemented")

        for asset, line in self.data.items():

            # print("DEPRECIATE",self.depreciation_dict)
            quantity = max(0, self.data[asset]["inventory"]*self.depreciation_dict[asset.strip()])

            if quantity > 0:
                self.remove_item(asset, quantity)

            else:
                error = NotImplementedError("Something went wrong. Tried to depreciate negative quantity.")
                raise error 


    def add_item(self, name, quantity, costs):
        """
        adds an item to the inventory

        :param name: name of the item
        :param quantity: quantity of the item (float or int)
        :param costs: the nominal value of the items being added (= quantity x price)
        """

        assert quantity >= 0

        self.data[name]["inventory"] += quantity
        self.data[name]["costs"] += costs

        self.owner.balance_sheet.disengage()
        self.owner.balance_sheet.change_item(name, BalanceEntry.ASSETS, costs, suppress_stock=True)
        self.owner.balance_sheet.change_item("Equity", BalanceEntry.EQUITY, costs, suppress_stock=True)
        self.owner.balance_sheet.engage()

    def remove_item(self, name, quantity):
        """
        Takes a certain quantity of an item out of inventory

        :param name: name of the item from which you want to remove
        :param quantity: physical quantity to remove
        """

        eps = 1e-10
        assert quantity <= self.data[name]["inventory"] + eps, "Exceeded inventory by" \
                                                         " removing a quantity (%.2f) more than on stock (%.2f)"%(quantity,self.data[name]["inventory"])

        avg_cost = self.data[name]["costs"]/self.data[name]["inventory"]

        """
        €/pieces * pieces = €
        """
        self.data[name]["inventory"] -= quantity
        self.data[name]["costs"] -= avg_cost * quantity

        self.owner.balance_sheet.disengage()
        self.owner.balance_sheet.change_item(name, BalanceEntry.ASSETS, -avg_cost * quantity, suppress_stock=True)
        self.owner.balance_sheet.change_item("Equity", BalanceEntry.EQUITY, -avg_cost * quantity, suppress_stock=True)

        # income statement: costs, e.g. costs for labor have already been subtracted
        # so there is no redundant income statement needed here

        self.owner.balance_sheet.engage()

        return avg_cost * quantity


    def get_inventory(self, name):
        """
        gets the current inventory of a certain item type

        :param: name of the item 
        """

        return self.data[name]["inventory"]

    @property
    def worth(self):
        """
        computes total nominal worth of the inventory.
        """

        s = 0
        for k, v in self.data.items():
            s += v["costs"]

        return s


__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '06.10.2022'
__status__ = 'dev' # options are: dev, test, prod

from sfctools.examples.example_wrapper import Example

from sfctools.datastructs.inventory import Inventory
from sfctools import Agent

def run():
    class InventoryAgent(Agent): # basic agent with an inventory
        def __init__(self):
            super().__init__()
            self.inventory = Inventory(self)

    # perform some arbitrary manipulations on the inventory
    my_agent = InventoryAgent()
    my_agent.inventory.add_item("Apple",1,3.0)
    my_agent.inventory.add_item("Apple",1,5.0)
    my_agent.inventory.remove_item("Apple", 1)
    w = my_agent.inventory.worth
    my_agent.inventory.add_item("Pear",2,2.0)
    my_agent.inventory.remove_item("Pear", 2)

    n_a = my_agent.inventory.get_inventory("Apple")
    n_p = my_agent.inventory.get_inventory("Pear")
    w2 = my_agent.inventory.worth

    print("n_a",n_a)
    print("n_p",n_p)
    print("w2",w2)


class InventoryExample(Example):
    def __init__(self):
        super().__init__(run)


if __name__ == "__main__":
    my_inv = InventoryExample()
    my_inv.run()


__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'dev' # options are: dev, test, prod


import numpy as np
from collections import defaultdict
import pandas as pd
from enum import Enum


class CashFlowEntry(Enum):
    """
    these are the different entry types
    """
    OPERATING = 0
    FINANCING = 1
    INVESTING = 2

class CashFlowStatement:
    """
    Cash flow statement data structure.
    See also https://www.investopedia.com/investing/what-is-a-cash-flow-statement/

    +------------------------------------+
    |    Cash Flow Statement             |
    +===============+====================+
    |  Operating    |                    |
    |  Activities   |                    |
    +---------------+--------------------+
    |               | Net Income | Value |
    +---------------+--------------------+
    |               |   Tag      | Value |
    +---------------+--------------------+
    |               |   Tag      | Value |
    +---------------+--------------------+
    |   TOTAL       |               ...  |
    +---------------+--------------------+
    |  Investing    |                    |
    |  Activities   |                    |
    +---------------+--------------------+
    |               |   Tag  |    Value  |
    +---------------+--------------------+
    |               |   Tag  |    Value  |
    +---------------+--------------------+
    |   TOTAL       |               ...  |
    +---------------+--------------------+
    |  Financing    |                    |
    |  Activities   |                    |
    |               |   Tag  |    Value  |
    +---------------+--------------------+
    |               |   Tag  |    Value  |
    +---------------+--------------------+
    |   TOTAL       |               ...  |
    +---------------+--------------------+


    """

    def __init__(self, owner):
        """
        instaniates a new cash flow statement

        :param owner: owner instance this cash flow statement belongs to.
        """

        self.owner = owner
        self.data = None
        self.tot_dict = None
        self.reset()

    def get_entry(self,kind,tag=None):
        """
        retrieve an arbitrary entry from the cash flow statement
        :param kind: sfctools.CashFlowEntry.OPERATING, FINANCING or INVESTING
        :param tag (optional): the tag under which to search for
        """
        if tag is None:
            return self.data[kind]
        else:
            return self.data[kind][tag]

    @property
    def total_cash_flow(self):
        """
        returns total cashflow, i.e. operating plus financing plus investing
        """
        return self.tot_dict[CashFlowEntry.OPERATING]  + self.tot_dict[CashFlowEntry.FINANCING] + self.tot_dict[CashFlowEntry.INVESTING]

    @property
    def operating_cash_flow(self):
        """
        returns only operating part of the cash flow
        """
        return self.tot_dict[CashFlowEntry.OPERATING]

    @property
    def financing_cash_flow(self):
        """
        returns only financing part of the cash flow
        """
        return self.tot_dict[CashFlowEntry.FINANCING]

    @property
    def investing_cash_flow(self):
        """
        returns only investing part of the cash flow
        """
        return self.tot_dict[CashFlowEntry.INVESTING]

    def new_entry(self, kind, tag, value):
        """
        registers a new item in the income statement

        :param kind: kind of entry: Financing, Operating or Investing
        :param tag: a name, e.g. 'interest on deposits'
        :param value: nominal value of the entry

        Example


        .. code-block:: python

            from sfctools import Agent, CashFlowEntry
            a = Agent()
            cfs = a.cash_flow_statement

            kind = CashFlowEntry.FINANCING
            tag = "My Subject"
            value = 10.0

            cfs.new_entry(kind,tag,value)
            print(cfs.to_dataframe())


        """

        self.data[kind][tag] += value
        self.tot_dict[kind] += value

    def reset(self):
        """
        Removes data and starts a new, blank income statement for the next period
        """

        self.data = {CashFlowEntry.OPERATING: defaultdict(lambda: 0.0),
                     CashFlowEntry.FINANCING: defaultdict(lambda: 0.0),
                     CashFlowEntry.INVESTING: defaultdict(lambda: 0.0),
                     }

        self.tot_dict = {
            CashFlowEntry.OPERATING: 0.0,
            CashFlowEntry.FINANCING: 0.0,
            CashFlowEntry.INVESTING: 0.0}

    def to_dataframe(self):
        """
        Converts the data structure to a human-readable pandas dataframe.
        """

        idx = "Value"

        keys = []

        df_operating = pd.DataFrame(self.data[CashFlowEntry.OPERATING], index=[idx]).T
        df_total_op = pd.DataFrame({"Tot. Operating": self.operating_cash_flow}, index=[idx]).T

        df_financing = pd.DataFrame(self.data[CashFlowEntry.FINANCING], index=[idx]).T
        df_total_fin = pd.DataFrame({"Tot. Financing": self.financing_cash_flow}, index=[idx]).T
        
        df_investing = pd.DataFrame(self.data[CashFlowEntry.INVESTING], index=[idx]).T
        df_total_inv = pd.DataFrame({"Tot. Investing": self.investing_cash_flow}, index=[idx]).T

        df_total = pd.DataFrame({"Total Cash Flow": self.total_cash_flow}, index=[idx]).T

        frames = [df_operating,df_total_op,df_financing,df_total_fin,df_investing,df_total_inv,df_total]

        df = pd.concat(frames,keys=["Operating","Tot. Operating", "Finacing","Tot. Financing","Investing","Tot. Investing","Tot."])

        df.index.names = ["Cash flow statement of %s" %self.owner,""]
        return df


    def to_string(self):
        """
        Converts the cash-flow statement to dataframe and then to string.
        """
        return self.to_dataframe().to_string()



__version__ = "0.3"
__author__ = "Thomas Baldauf"
__email__ = "thomas.baldauf@dlr.de"
__license__ = "MIT"
__birthdate__ = '15.11.2021'
__status__ = 'test' # options are: dev, test, prod

import pandas as pd
import sys 

sys.path.append("..")


from ..core.singleton import Singleton
from ..core.clock import Clock
from ..datastructs.collection import Collection
import numpy as np



class StockManager(Singleton):
    """
    The stock manager keeps track of all price trends. 
    Beta feature...
    """

    def __init__(self):
        """
        constructor of stock manager.
        """

        if hasattr(self, "initialized"):  # only initialize once
            return
        self.initialized = True

        self.hist = Collection(kind="list", maxlen=4096)  # history  logger with maximum length
        self.state = {}        # keys are price index, values are current stock values
        self.abbrev_dict = {}  # keys are commodity name, values are price index names

    def register(self, name: str, price_name: str, price_value: float) -> None:
        """
        run setup routine of stock manager for a new commodity
        
        :param name: name of the commodity  (eg. 'coal')
        :param price_name: name of the price  (e.g. 'P_c')
        :param price_value: actual value of the price, this can also point to a method
        """

        self.abbrev_dict[name] = price_name
        self.state[price_name] = price_value

        self.set_price(price_name,price_value)

    def get_index(self, name_of_commodity: str) -> str:
        """
        retrieve stock index by name of commodity

        :param name_of_commodity: name of the commodity
        """
        return self.abbrev_dict[name_of_commodity]

    def get_price(self, name, human=False):
        """
        see request_price
        """
        self.request_price(name, human)

    def get_most_recent(self, name):
        """
        get the most recent value of a price index
        if there are no values, return None
       
        :return: t, v (time stamp and value of most recent update)
        """

        entries = self.hist.get_entries()
        if name in entries:
            times, values = np.array(entries[name]).T
            return times[-1], values[-1]

        else:
            return None

    def request_price(self, name: str, human: bool = False) -> float:
        """
        Retrieve a price by the name of the index.

        :param name: the name of the commodity, e.g. 'coal'
        :param human: get price by the name of the commodity instead
        :return:  P, value of the price at the stock
        """

        # NOTE: the name of this method is request_price, not get_price because
        # in the future we want to add access rights of agents to certain price indices
        # this could also include fuzzy information propagation or just rejection to access a certain index,
        # possibly at random... so the request might be also denied in the future

        if human:
            key = self.abbrev_dict[name]
        else:
            key = name

        return self.state[key]

    def set_price(self, name: str, value: float, human: bool = False) -> None:
        """
        set a new price by name of the index commodity
        
        :param name:  name of the index to be set
        :param value: new value
        :param human: get price by the name of the commodity instead
        
        """

        # 1. log the old value

        if human:
            key = self.abbrev_dict[name]
        else:
            key = name

        assert key in self.abbrev_dict or key in self.state, "Price index %s not known" % name

        self.hist.insert(key, (Clock().get_time(), value))

        # 2. update the new value

        self.state[key] = value

    def get_info(self) -> pd.DataFrame:
        """
        get stock information in a table

        :return: df, pandas dataframe
        """

        # TODO optimize formatting for nice printing

        df_data = {"Commodity": [], "Price Index": [], "Current Stock Value": []}

        for key, val in self.abbrev_dict.items():
            p_val = self.state[val]
            df_data["Current Stock Value"].append(p_val)
            df_data["Price Index"].append(val)
            df_data["Commodity"].append(key)

        df = pd.DataFrame(df_data)

        df.set_index("Commodity", inplace=True)
        return df

    def request_price_history(self, name: str, which: str = "all", convert_df: bool = False, convert_dict: bool = False, human: bool = False):
        """
        get historical trace of a stock index 

        :param name: name of the commodity
        :param which: passing 'all' (default) will return all past values, passing 'last' will only return the two latest values. BETA works only in convert_dict mode
        :param convert_df: convert to dataframe? if False, a numpy array is returned instead (which is faster)
        :param convert_dict: convert to dictionary, where keys are time periods and values are the data values; the numpy array will be of shape [2,n] where [0,0...n-1] is the time stamp and [1,0....n-1] is the price data belonging to this  time stamp
        :param human: get price history by the name of the commodity instead
        :return: df or dict

        Example

        .. code-block:: python

            from sfctools import Clock, StockManager
            import numpy as np 

            for i in range(10):
                StockManagre().set_price("P_K",10.0 + np.random.rand())
                Clock().tick()

            times, values = StockManager().request_price_history("P_K")
            print(times)
            print(values)

        """

        if human:
            key = self.abbrev_dict[name]
        else:
            key = name

        values = np.array(self.hist.get_entries()[key])  # get data

        if not convert_df:
            if not convert_dict:
                return values.T
            else:

                if len(values) == 0:
                    return {}

                times, vals = values.T

                d = {}

                N = len(times)
                if which== "last":
                    N = min(3,len(times))
                    # get values for t, t-1, t-2 which are typically required in dynamic macro models
                    # (Bellman equation kicks out rest)

                for i in range(len(times)-N,len(times)):
                    # d[times[N-1-i]] = vals[N-1-i]
                    d[times[i]] = vals[i]

                return d
        else:

            df = pd.DataFrame(values, columns=["Time", key]).set_index("Time")
            return df
